#!/usr/bin/env python
import sys
import os

from matplotlib import transforms
from matplotlib.colors import get_named_colors_mapping
import matplotlib.pyplot as plt
from matplotlib.pyplot import figure, get
import numpy as np
from matplotlib.patches import Circle
from pandas.core import algorithms
from scipy import cluster

from .utils import read_file,save_seqs
from .character import Character
from .column import Column
from .item import Item
from .utils import get_coor_by_angle,  link_edges, rotate, text3d
from .connect import get_connect, get_score_mat, msa
from matplotlib.patches import Circle, PathPatch
from matplotlib.text import TextPath
from matplotlib.transforms import Affine2D
import mpl_toolkits.mplot3d.art3d as art3d
import seaborn as sns
from matplotlib.lines import Line2D
from matplotlib.patches import Arc, RegularPolygon
from numpy import radians as rad
import math
import re
import time
import pandas as pd
import pathlib
import os
from scipy.spatial import distance
from scipy.cluster import hierarchy
from scipy.cluster.hierarchy import dendrogram, linkage
from mpl_toolkits.axes_grid1 import make_axes_locatable



from .utils import grouping,detect_seq_type
from .logobits import compute_bits, compute_prob
from .colors import get_color_scheme
from .version import __version__

basic_dna_color = get_color_scheme('basic_dna_color')


class Logo(Item):
    def __init__(self, bits, ax = None, start_pos=(0,0), logo_type='Horizontal', column_width=1, 
                 column_margin_ratio=0.1, char_margin_ratio = 0.1, parent_start = (0,0), origin = (0,0), id='', 
                 help_color='b', color=basic_dna_color, limited_char_width=None, path_dict={}, *args, **kwargs):
        super(Logo, self).__init__(*args, **kwargs)

        self.bits = bits
        self.start_pos = start_pos
        self.logo_type = logo_type
        self.parent_start = parent_start
        self.column_margin_ratio = column_margin_ratio
        self.char_margin_ratio = char_margin_ratio
        self.column_width = column_width
        self.origin = origin
        self.id = id
        self.color = color
        self.help_color = help_color
        self.columns = []
        self.limited_char_width = limited_char_width
        self.path_dict = path_dict

        if ax == None:
            self.ax = self.generate_ax(threed=(self.logo_type=='Threed'))
        else:
            self.ax = ax

        if limited_char_width == None:
            self.limited_char_width = self.get_limited_char_width()
        
        self.generate_components()

    def generate_components(self):
        for index,bit in enumerate(self.bits):
            chars = [x[0] for x in bit]
            weights = [x[1] for x in bit]
            if chars == []:
                chars = ['-']
                weights = [0]
            column = Column(chars,weights,ax=self.ax,width=self.column_width,logo_type=self.logo_type,
                            origin=self.origin, color=self.color, char_margin_ratio=self.char_margin_ratio,
                            limited_char_width=self.limited_char_width,path_dict=self.path_dict)
            self.columns.append(column)
    
    def draw(self):
        self.compute_positions()
        for col in self.columns:
            col.draw()
    
    def draw_help(self,show_id=True,group_id_size=10, **kwargs):

        if self.logo_type == 'Threed': 
            self.draw_3d_help(show_id=show_id, group_id_size=group_id_size, **kwargs)
       
        if self.logo_type == 'Horizontal': 
            self.draw_hz_help(show_id=show_id, group_id_size=group_id_size,**kwargs)
        
        if self.logo_type == 'Circle':
            self.draw_circle_help(show_id=show_id, group_id_size=group_id_size,**kwargs)
        
        if self.logo_type == 'Radiation':
            self.draw_rad_help(show_id=show_id, group_id_size=group_id_size,**kwargs)
    
    def draw_rad_help(self, show_id=True, group_id_size=10, **kwargs):
        if show_id:
            label_radius = (self.start_pos[0] + self.get_width() ) 
            label_x = label_radius * np.cos(self.deg)
            label_y = label_radius * np.sin(self.deg)
            self.id_txt = self.ax.text(label_x,label_y, f'{self.id}',rotation=math.degrees(self.deg),fontsize=group_id_size)
    
    def draw_hz_help(self,show_id=True,group_id_size=10, **kwargs):
        if show_id:
            #self.id_txt = self.ax.text(self.get_width() + 0.5, self.start_pos[1]+0.1,
            self.id_txt = self.ax.text(self.get_width() + 0.5, self.start_pos[1]+self.get_height()*0.1,
                                     f"{self.id}", fontsize=group_id_size, clip_on=True)#,bbox={'fc': '0.8', 'pad': 0})
        

        
    def draw_circle_help(self,show_id=True, group_id_size=10,draw_arrow=False,**kwargs):
        self.ax.add_patch(Circle(self.parent_start,self.radius,linewidth=1,fill=False,edgecolor='grey',alpha=0.5))

        space_deg = self.degs[0] + (self.degs[-1] - self.degs[0])/2
        space_coor = get_coor_by_angle(self.radius ,space_deg)
        self.ax.scatter(space_coor[0],space_coor[1],color=self.help_color)

        space_coor2 = get_coor_by_angle(self.radius + self.get_height() ,space_deg)
        #self.ax.plot([self.parent_start[0],space_coor[0]],[self.parent_start[1],space_coor[1]],zorder=-1)

        self.ax.plot([space_coor[0],space_coor2[0]],[space_coor[1],space_coor2[1]],zorder=-1,color='grey')

        if draw_arrow == True:
            self.ax.plot([self.origin[0],space_coor[0]],[self.origin[1],space_coor[1]],zorder=-1,color='grey')
            arc = Arc(self.origin,self.radius,self.radius,angle=270,
                  theta1=0,theta2=180,capstyle='round',linestyle='-',lw=2,color='black')
            self.ax.add_patch(arc)
            endX = 0
            endY = -self.radius/2
            self.ax.add_patch(                    #Create triangle as arrow head
                 RegularPolygon(
                    (endX, endY),            # (x,y)
                     3,                       # number of vertices
                    self.radius/9,                # radius
                    rad(30+180),     # orientation
                    color='black'
                )
            )

    
    def draw_3d_help(self,z_height_3d=2, show_id=True, group_id_size=10,**kwargs):
        if show_id:
            self.ax.text(0, self.start_pos[2], z_height_3d, f'{self.id}', 'z',fontsize=group_id_size)



    
    def compute_positions(self):

        if self.logo_type == 'Circle':
            self.column_margin_ratio = 0
            self.radius = np.sqrt((self.start_pos[0]-self.parent_start[0])**2 + (self.start_pos[1]-self.parent_start[1])**2)
            self.each_deg = 2*np.pi / len(self.bits)
            width = 2 * self.radius * np.tan(self.each_deg/2)
            width = width * 0.95
            self.column_width = width
            degs = [x*self.each_deg + np.pi/2  for x in range(len(self.bits))]
            degs = degs[::-1]
            degs = [degs.pop()] + degs
            self.degs = degs

        start_pos = self.start_pos
        for index,col in enumerate(self.columns):
            col.set_parent_start(self.start_pos)
            if self.logo_type == 'Horizontal':
                col.set_start_pos(start_pos)
                col.compute_positions()
                start_pos = (start_pos[0] + col.get_width() * (1+self.column_margin_ratio), start_pos[1])
            elif self.logo_type == 'Circle':
                start_pos = get_coor_by_angle(self.radius, self.degs[index], self.parent_start)
                col.set_start_pos(start_pos)
                col.set_deg(self.degs[index])
                col.set_width(self.column_width)
                col.compute_positions()
            elif self.logo_type == 'Radiation':
                col.set_start_pos(start_pos)
                col.set_deg(self.deg)
                col.set_radiation_space(self.radiation_space)
                col.compute_positions()
                start_pos = (start_pos[0] + col.get_width() *(1+self.column_margin_ratio), start_pos[1])
            elif self.logo_type == 'Threed':
                col.set_start_pos(start_pos)
                col.compute_positions()
                start_pos = (start_pos[0] + col.get_width() *(1+self.column_margin_ratio), start_pos[1], start_pos[2])
            else:
                pass
    
    def get_height(self):
        return max([col.get_height() for col in self.columns]+[0])
    
    def get_width(self):
        return sum([col.get_width() *(1+self.column_margin_ratio) for col in self.columns])




class LogoGroup(Item):
    def __init__(self,  seqs, ax=None, group_order='length', group_strategy='length', group_resolution=0.5,
                 clustering_method = 'max', min_length = 0, max_length = 100,
                 start_pos = (0,0), logo_type = 'Horizontal', init_radius=1, 
                 logo_margin_ratio = 0.1, column_margin_ratio = 0.05, char_margin_ratio = 0.05,
                 align = True, align_metric='sort_consistency', connect_threshold=0.8, 
                 radiation_head_n = 5, threed_interval = 4, color = basic_dna_color, task_name='MetaLogo',
                 x_label = 'Position', y_label = 'bits',z_label = 'bits', show_grid = True, show_group_id = True,
                 display_range_left = 0, display_range_right = -1,
                 hide_left_axis=False, hide_right_axis=False, hide_top_axis=False, hide_bottom_axis=False,
                 hide_x_ticks=False, hide_y_ticks=False, hide_z_ticks=False, 
                 title_size=20, label_size=10, tick_size=10, group_id_size=10,align_color='blue',align_alpha=0.1,
                 figure_size_x=-1, figure_size_y=-1,gap_score=-1, padding_align=False, hide_version_tag=False,
                 sequence_type = 'auto', height_algorithm = 'bits',omit_prob = 0,
                 seq_file = '', seq_file_type = 'fasta', fa_output_dir = '.', output_dir = '.', uid = '',
                 withtree = False,group_limit=20, target_sequence_name = '',
                 clustalo_bin = '', fasttree_bin = '', fasttreemp_bin = '', treecluster_bin = '',
                 auto_size=True,
                 *args, **kwargs):
        super(LogoGroup, self).__init__(*args, **kwargs)
        self.seqs = seqs
        self.seq_file = seq_file
        self.seq_file_type = seq_file_type
        self.min_length = 0
        self.max_length = 100
        self.target_sequence_name = target_sequence_name
        self.target_sequence_content = None
        self.group_order = group_order
        self.group_strategy = group_strategy
        self.group_resolution = float(group_resolution)
        self.start_pos = start_pos
        self.logo_margin_ratio = logo_margin_ratio
        self.column_margin_ratio = column_margin_ratio
        self.char_margin_ratio = char_margin_ratio
        self.logo_type = logo_type
        self.init_radius = init_radius
        self.radiation_head_n = 5
        self.threed_interval = threed_interval
        self.align = align 
        self.ceiling_pos = (0,1)
        self.align_metric = align_metric
        self.connect_threshold = connect_threshold
        self.color = color
        self.task_name = task_name

        self.height_algorithm = height_algorithm
        self.omit_prob = omit_prob

        self.align_color = align_color
        self.align_alpha = align_alpha

        self.padding_align = padding_align

        self.display_range_left = display_range_left
        self.display_range_right = display_range_right

        self.gap_score = gap_score

        self.hide_left_axis = hide_left_axis
        self.hide_right_axis = hide_right_axis
        self.hide_bottom_axis = hide_bottom_axis
        self.hide_top_axis = hide_top_axis

        self.hide_x_ticks = hide_x_ticks
        self.hide_y_ticks = hide_y_ticks
        self.hide_z_ticks = hide_z_ticks

        self.x_label = x_label
        self.y_label = y_label
        self.z_label = z_label

        self.tick_size = tick_size
        self.title_size = title_size
        self.label_size = label_size
        self.group_id_size = group_id_size

        self.show_group_id = show_group_id
        self.show_grid = show_grid

        self.figure_size_x = figure_size_x
        self.figure_size_y = figure_size_y

        self.hide_version_tag = hide_version_tag

        self.clustalo_bin = clustalo_bin
        self.fasttree_bin = fasttree_bin
        self.fasttreemp_bin = fasttreemp_bin
        self.treecluster_bin = treecluster_bin

        self.fa_output_dir = fa_output_dir
        self.output_dir = output_dir
        self.uid = uid

        self.clustering_method = clustering_method
        self.withtree = withtree
        
        self.group_limit = group_limit
        self.auto_size = auto_size

        if (self.seqs is None) and (not os.path.exists(self.seq_file)):
            print('No sequences provided')
            self.error = 'No sequences detected'
            return

        if (self.seqs is None) and os.path.exists(self.seq_file):
            seq_dict,seqnames = read_file(self.seq_file, self.seq_file_type)

            if len(seqnames) == 0:
                print('No sequences detected')
                self.error = 'No sequences detected'
                return

            if (len(seq_dict[seqnames[0]]) < self.min_length) or (len(seq_dict[seqnames[0]]) > self.max_length):
                print('The first sequence not satisfied the length limit')
                self.error = 'The first sequence not satisfied the length limit'
                return

            seqs = [[seqname,seq_dict[seqname]]  for seqname in seqnames if (len(seq_dict[seqname])>self.min_length) and (len(seq_dict[seqname])<self.max_length)]
            target_sequence_name = ' ' .join(seqnames[0].split(' ')[:-1])

            if len(seqs) == 0:
                print('No sequences left after length filter')
                self.error = 'No sequences left after length filter'
                return
            
            self.seqs = seqs
            if self.target_sequence_name == '':
                self.target_sequence_name = target_sequence_name
        
        if self.seqs is not None:
            if self.target_sequence_name == '':
                self.target_sequence_name = self.seqs[0][0]
        
        if self.seqs is not None:
            if not os.path.exists(self.seq_file):
                if self.seq_file != '':
                    save_seqs(self.seqs,self.seq_file)
                else:
                    save_seqs(self.seqs,f'{self.fa_output_dir}/server.{self.uid}.fasta')

        if self.seq_file_type.lower() in ['fastq','fq']:
            save_seqs(self.seqs,f'{self.fa_output_dir}/server.{self.uid}.fasta')
            self.seq_file = f'server.{self.uid}.fasta'

        if sequence_type == 'auto':
            self.sequence_type = detect_seq_type(self.seqs)
        else:
            self.sequence_type = sequence_type
        
        self.check_dep()
        if hasattr(self,'error'):
            return

        self.logos = []

        self.prepare_bits()
        if hasattr(self,'error'):
            return

        if ax is None:
            withtree = False
            if (len(self.group_ids) > 1) and (self.withtree) and (self.logo_type == 'Horizontal') and ( self.group_strategy == 'auto' or (self.align and self.padding_align)):
                withtree = True
            self.generate_ax(threed=(self.logo_type=='Threed'),withtree=withtree)
        else:
            self.ax = ax


        self.generate_components()
    
    def check_dep(self):
        if not os.path.exists(self.clustalo_bin): 
            err = 'Clustal omega not found'
            print(err)
            self.error = err
        elif not os.path.exists(self.fasttree_bin):
            err = 'FastTree not found'
            print(err)
            self.error = err
        elif not os.path.exists(self.fasttreemp_bin):
            err = 'FastTreeMP not found'
            print(err)
            self.error = err
        else:
            pass
        return

    
    def prepare_bits(self):
        self.groups = grouping(self.seqs,seq_file=self.seq_file,sequence_type=self.sequence_type,group_by=self.group_strategy,
                               group_resolution=self.group_resolution,clustering_method=self.clustering_method,
                               clustalo_bin=self.clustalo_bin,fasttree_bin=self.fasttree_bin,fasttreemp_bin=self.fasttreemp_bin,treecluster_bin=self.treecluster_bin,
                               uid=self.uid,fa_output_dir=self.fa_output_dir,figure_output_dir=self.output_dir)
        
        self.raw_group_count = len(self.groups)

        self.target_group = None
        for grpid in self.groups:
            for name,seq in self.groups[grpid]:
                if name == self.target_sequence_name:
                    self.target_group = grpid
                    self.target_sequence_content = seq
                    break
            if self.target_group is not None:
                break

        if self.group_strategy.lower() == 'identifier':
            for group_id in self.groups:
                seqs = self.groups[group_id]
                if len(set([len(x[1]) for x in seqs])) > 1:
                    print('Sequence lengths not same in one group')
                    self.error = 'In identifier-grouping mode, sequence lengths should be the same in one group'
                    return
        
        if len(self.groups) > self.group_limit :
            new_groups = {}
            sorted_groups = sorted(self.groups.items(),key=lambda d:len(d[1]),reverse=True)
            for gid,group in sorted_groups[:max(0,self.group_limit-1)]:
                new_groups[gid] = group
            if self.target_group is not None:
                new_groups[self.target_group] = self.groups[self.target_group]
            else:
                if self.group_limit > 0 :
                    add_grp_id,add_grp = sorted_groups[self.group_limit-1]
                    new_groups[add_grp_id] = add_grp
            self.groups = new_groups
        

        self.probs = compute_prob(self.groups,threshold=self.omit_prob)

        if self.height_algorithm == 'probabilities':
            self.seq_bits = self.probs.copy()
            seq_bits = {}
            for key in self.probs:
                seq_bits[key] = []
                for pos in self.probs[key]:
                    item = []
                    for base,height in pos:
                        if base != '-':
                            item.append((base,height))
                    seq_bits[key].append(item)
            self.seq_bits = seq_bits

        elif self.height_algorithm in ['bits','bits_without_correction']:
            self.seq_bits = compute_bits(self.groups, self.probs, seq_type=self.sequence_type,no_correction=self.height_algorithm=='bits_without_correction')
        
        try:
            if self.group_order.lower() == 'length':
                self.group_ids = sorted(self.seq_bits.keys(),key=lambda d: len(self.seq_bits[d]))
            elif self.group_order.lower() == 'length_reverse':
                self.group_ids = sorted(self.seq_bits.keys(),key=lambda d: len(self.seq_bits[d]),reverse=True)
            elif self.group_order.lower() == 'identifier':
                self.group_ids = sorted(self.seq_bits.keys(),key=lambda d: re.split('[@-]',d)[0])
            elif self.group_order.lower() == 'identifier_reverse':
                self.group_ids = sorted(self.seq_bits.keys(),key=lambda d: re.split('[@-]',d)[0], reverse=True)
            elif self.group_order.lower() == 'auto':
                self.group_ids = list(self.seq_bits.keys())
        except Exception as e:
            print(e)
            self.group_ids = sorted(self.seq_bits.keys())
        
        if self.height_algorithm in ['bits','bits_without_correction']:
            to_del_ids = []
            for gid in self.group_ids:
                total = sum([sum([x[1] for x in col]) for col in self.seq_bits[gid]])
                if total == 0:
                    to_del_ids.append(gid)

            for gid in to_del_ids: 
                self.group_ids.remove(gid)
                del self.probs[gid]
                del self.seq_bits[gid]
        

        if (self.align and self.padding_align) and (self.group_strategy != 'auto'):

            if len(self.group_ids) > 1:
                if self.align_metric in ['js_divergence','entropy_bhattacharyya']:
                    probs_list = [self.probs[gid] for gid in self.group_ids]
                    self.scores_mat = get_score_mat(probs_list,align_metric=self.align_metric,gap_score=self.gap_score,seq_type=self.sequence_type)
                    new_probs_list = msa(probs_list,self.scores_mat,align_metric=self.align_metric,gap_score=self.gap_score,seq_type=self.sequence_type)
                    self.probs = dict(zip(self.group_ids, new_probs_list))
                else:
                    seq_bits_list = [self.seq_bits[gid] for gid in self.group_ids]
                    self.scores_mat = get_score_mat(seq_bits_list,align_metric=self.align_metric,gap_score=self.gap_score,seq_type=self.sequence_type)
                    new_seq_bits_list = msa(seq_bits_list,self.scores_mat,align_metric=self.align_metric,gap_score=self.gap_score,seq_type=self.sequence_type)
                    self.seq_bits = dict(zip(self.group_ids, new_seq_bits_list))
            
                self.align_probs_bits()
        
        
        if self.padding_align or (self.group_strategy == 'auto'):
            if not (self.display_range_left == 0 and self.display_range_right == -1):

                self.task_name = f"{self.task_name} (Display range: {self.display_range_left} to {self.display_range_right})"

                new_probs = {}
                new_bits = {}

                for gid in self.group_ids:
                    new_probs[gid] = self.probs[gid][self.display_range_left: self.display_range_right + 1*(self.display_range_right!=-1)]
                    new_bits[gid] = self.seq_bits[gid][self.display_range_left: self.display_range_right + 1*(self.display_range_right!=-1)]
            
                self.full_probs = self.probs.copy()
                self.full_seq_bits = self.seq_bits.copy()
            
                self.probs = new_probs
                self.seq_bits = new_bits
            
            ##get correlations
            if len(self.group_ids) > 1:
                self.correlation  = self.get_correlation()
                correlation_array = np.asarray(self.correlation)
                self.row_linkage = hierarchy.linkage(distance.pdist(correlation_array.T), method='average')
                self.dendrogram = dendrogram(self.row_linkage,orientation='left')
                self.before_clustering_group_ids = self.group_ids.copy()
                self.group_ids = list(self.correlation.index[self.dendrogram['leaves']])

                

    
    def align_probs_bits(self):
        new_probs = {}
        new_bits = {}
        for gid in self.group_ids:
            new_probs[gid] = []
            new_bits[gid] = []
            prob_i = 0
            bit_i = 0
            while prob_i < len(self.probs[gid]) and bit_i < len(self.seq_bits[gid]):
                if self.probs[gid][prob_i] == []:
                    new_probs[gid].append([])
                    new_bits[gid].append([])
                    prob_i += 1
                elif self.seq_bits[gid][bit_i] == []:
                    new_probs[gid].append([])
                    new_bits[gid].append([])
                    bit_i += 1
                else:
                    new_probs[gid].append(self.probs[gid][prob_i].copy())
                    new_bits[gid].append(self.seq_bits[gid][bit_i].copy())
                    bit_i += 1
                    prob_i += 1

            max_len = max(len(self.probs[gid]),len(self.seq_bits[gid]))
            
            for i in range(len(new_probs[gid]),max_len):
                new_probs[gid].append([])
            for j in range(len(new_bits[gid]),max_len):
                new_bits[gid].append([])

        self.probs = new_probs
        self.seq_bits = new_bits


    def generate_components(self):


        self.help_color_palette = sns.color_palette("hls", len(self.seq_bits))

        self.limited_char_width = self.get_limited_char_width()
        self.generate_all_path()

        for index,group_id in enumerate(self.group_ids):
            bits = self.seq_bits[group_id]
            logo = Logo(bits,ax=self.ax,logo_type=self.logo_type,parent_start=self.start_pos,
                        origin=self.start_pos,id=group_id,
                        help_color=self.help_color_palette[index], color=self.color,
                        column_margin_ratio=self.column_margin_ratio, char_margin_ratio=self.char_margin_ratio,
                        limited_char_width=self.limited_char_width, path_dict=self.path_dict)
            self.logos.append(logo)

    def generate_all_path(self):
        path_dict = {}
        for base in 'ABCDEFGHIJKLMNOPQRSTUVWXYZ':
            tmp_path = TextPath((0,0), base, size=1)
            bbox = tmp_path.get_extents()
            path_dict[base] = [tmp_path, bbox]
        self.path_dict = path_dict

    def set_font(self):
        pass
    
    def draw(self):
        if hasattr(self,'error'):
            print('error: ', self.error)
            return
        self.compute_positions()

        z_height_3d = max([logo.get_height() for logo in self.logos]+[0])
        for index,logo in enumerate(self.logos):
            logo.draw()
            logo.draw_help(draw_arrow=index==0,z_height_3d=z_height_3d, show_id=self.show_group_id,
                            group_id_size=self.group_id_size)
        
        #draw connect


        if self.align:
            self.draw_connect()


        self.draw_help()
        self.compute_xy()
        self.set_figsize()
        if self.group_strategy == 'auto':
            task_name = self.task_name + '\n-' + '[%s resolution]'%(self.group_resolution)
        else:
            task_name = self.task_name + '\n-'
        if hasattr(self,'raw_group_count'):
            task_name += f' [{len(self.groups)}/{self.raw_group_count} groups]' 
        if self.group_strategy == 'auto' or (self.align and self.padding_align ):
            if len(self.group_ids)>0:
                poses = len(self.seq_bits[self.group_ids[0]])
                left = self.display_range_left
                right = self.display_range_right
                if right == -1:
                    right = poses
                task_name += f' [{left}:{right} range]' 
        
        
        self.ax.set_title(task_name,fontsize=self.title_size,loc='left')

        self.ax.tick_params(labelsize=self.tick_size)


        self.ax.spines['left'].set_visible(not self.hide_left_axis)
        self.ax.spines['right'].set_visible(not self.hide_right_axis)
        self.ax.spines['top'].set_visible(not self.hide_top_axis)
        self.ax.spines['bottom'].set_visible(not self.hide_bottom_axis)


        if  self.hide_x_ticks:
            if self.show_grid:
                self.ax.set_xticklabels([])
            else:
                self.ax.set_xticks([])

        if  self.hide_y_ticks:
            if self.show_grid:
                self.ax.set_yticklabels([])
            else:
                self.ax.set_yticks([])
        if self.hide_z_ticks and self.logo_type == 'Threed':
            if self.show_grid:
                self.ax.set_zticklabels([])
            else:
                self.ax.set_zticks([])
        
        if self.show_grid:
            self.ax.grid(True)
        
        self.ax.set_xlabel(self.x_label,fontsize=self.label_size) 
        self.ax.set_ylabel(self.y_label,fontsize=self.label_size) 

        if self.logo_type == 'Threed':
            self.ax.set_zlabel(self.z_label,fontsize=self.label_size)
        
        if not self.hide_version_tag:
            if self.logo_type == 'Threed':
                #self.ax.text2D(1.005, 0, 'Created by MetaLogo')
                self.ax.text2D(1.00, 0, f'Created by MetaLogo (v{__version__})', transform=self.ax.transAxes,
                    horizontalalignment='left',
                    verticalalignment='bottom',
                    rotation='vertical',
                    color='#6c757d')

            else:
                #self.ax.text(1.005, 0, f'Created by MetaLogo (v{__version__})',
                self.ax.text(1.005, 1, f'Created by MetaLogo (v{__version__})',
                    #horizontalalignment='left',
                    horizontalalignment='right',
                    #verticalalignment='bottom',
                    verticalalignment='bottom',
                    #rotation='vertical',
                    transform=self.ax.transAxes,
                    color='#6c757d')
        
        if (len(self.group_ids) > 1) and (self.withtree) and (self.logo_type == 'Horizontal') and (self.group_strategy=='auto' or (self.padding_align and self.align)):
            #draw new tree
            intervals = []
            total_height = 0
            for logo in self.logos:
                h = logo.get_height()
                intervals += [total_height + h/2]
                total_height += h*(1+self.logo_margin_ratio)
            Zx = []
            Zy = []
            cluster_center = {}
            i = -1
            for item in self.row_linkage:
                i += 1
                node1,node2,branch,num = item
                node1 = int(node1)
                node2 = int(node2)

                if node1 < len(self.group_ids) and node2 < len(self.group_ids):
                    idx1 = self.group_ids.index(self.before_clustering_group_ids[node1])
                    idx2 = self.group_ids.index(self.before_clustering_group_ids[node2])
                    if idx1 > idx2:
                        idx1,idx2 = idx2,idx1
                    if num == 2:
                        p1 = [0, intervals[idx1]]
                        p2 = [branch,intervals[idx1]]
                        p3 = [branch,intervals[idx2]]
                        p4 = [0, intervals[idx2]]
                        Zx += [[p1[0],p2[0],p3[0],p4[0]]]
                        Zy += [[p1[1],p2[1],p3[1],p4[1]]]
                        cluster_center[len(self.group_ids)+i] = [branch,p1[1] + (p4[1]-p1[1])/2]
                else:
                    if node1 >= len(self.group_ids):
                        p1 = cluster_center[node1]
                    else:
                        idx1 = self.group_ids.index(self.before_clustering_group_ids[node1])
                        p1 = [0,intervals[idx1]]

                    if node1 >= len(self.group_ids):
                        p1 = cluster_center[node1]
                    else:
                        idx1 = self.group_ids.index(self.before_clustering_group_ids[node1])
                        p1 = [0,intervals[idx1]]
                    if node2 >= len(self.group_ids):
                        p4 = cluster_center[node2]
                    else:
                        idx2 = self.group_ids.index(self.before_clustering_group_ids[node2])
                        p4 = [0,intervals[idx2]]
                    
                    if p1[1] > p4[1]:
                        p1,p4 = p4,p1
                    
                    p2 = [branch,p1[1]]
                    p3 = [branch,p4[1]]

                    Zx += [[p1[0],p2[0],p3[0],p4[0]]]
                    Zy += [[p1[1],p2[1],p3[1],p4[1]]]
                    cluster_center[len(self.group_ids)+i] = [branch,p1[1]+(p4[1]-p1[1])/2]
                
            for path_x,path_y in zip(Zx,Zy):
                self.ax0.plot(path_x,path_y)

            if self.target_group is not None:
                if self.target_group in self.group_ids:
                    target_y = intervals[self.group_ids.index(self.target_group)]
                    self.ax0.plot(0,target_y,'ro',color='red')
            
            self.ax0.invert_xaxis()
            self.ax0.spines['left'].set_visible(False)
            self.ax0.spines['right'].set_visible(False)
            self.ax0.spines['top'].set_visible(False)
            self.ax0.spines['bottom'].set_visible(False)
            self.ax0.set_yticks([])
            self.ax0.set_xticks([])
            self.ax0.set_ylim(self.ax.get_ylim())

            self.ax.yaxis.tick_right()
            self.ax.yaxis.set_label_position("right")


    def draw_help(self):
        if self.logo_type == 'Radiation':
            self.draw_radiation_help()
        
        if self.logo_type == 'Circle':
            self.draw_circle_help()
        
        if self.logo_type == 'Horizontal':
            self.draw_hz_help()
    
    def draw_connect(self):
        if self.align_metric in ['js_divergence','entropy_bhattacharyya']:
            self.connected = get_connect([self.probs[gid] for gid in self.group_ids], self.align_metric, 
                                       gap_score=self.gap_score,msa_input= (self.padding_align or self.group_strategy=='auto'),seq_type=self.sequence_type)
        else:
            self.connected = get_connect([self.seq_bits[gid] for gid in self.group_ids], self.align_metric, 
                                       gap_score=self.gap_score,msa_input=(self.padding_align or self.group_strategy=='auto'),seq_type=self.sequence_type)
        if self.connect_threshold < 0:
            vals = []
            for group_id in self.connected:
                for pos,arr in self.connected[group_id].items():
                    vals.append(arr[0])
            filtered_sorted_vals = sorted(vals, reverse=True)[:int(len(vals)*abs(self.connect_threshold))]
            if len(filtered_sorted_vals) > 0:
                self.connect_threshold = filtered_sorted_vals[-1]
            else:
                self.connect_threshold = 1E9

        i = -1
        for group_id  in self.group_ids:
            i += 1
            if group_id == self.group_ids[-1]:
                continue
            link = self.connected[i]
            for pos1, arr in link.items():
                r, targets = arr
                if r > self.connect_threshold:
                    for pos2 in targets:
                        self.link_columns(self.logos[i].columns[pos1], self.logos[i+1].columns[pos2])


    
    def link_columns(self, column1, column2, ):

        nodes1 = column1.get_edge()
        nodes2 = column2.get_edge()

        if self.logo_type == 'Threed':
            link_edges((nodes1[0],nodes1[1]), (nodes2[0],nodes2[1]) , self.ax, threed=True, x=0,y=2,z=1, 
                        color=self.align_color, alpha=self.align_alpha)
        else:
            if self.logo_type == 'Radiation':
                link_edges((nodes1[0],nodes1[1]), (nodes2[3],nodes2[2]) , self.ax,
                            color=self.align_color, alpha=self.align_alpha)
            else:
                link_edges((nodes1[3],nodes1[2]), (nodes2[0],nodes2[1]) , self.ax,
                            color=self.align_color, alpha=self.align_alpha)
            link_edges((nodes2[0],nodes2[1]), (nodes2[3],nodes2[2]) , self.ax,
                        color=self.align_color, alpha=self.align_alpha)
            link_edges((nodes1[0],nodes1[1]), (nodes1[3],nodes1[2]) , self.ax,
                        color=self.align_color, alpha=self.align_alpha)


    def draw_circle_help(self):           
        #https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.legend.html
        if self.show_group_id:
            legend_elements = []
            for logo in self.logos[::-1]:
                legend_elements.append( Line2D([0], [0], marker='o', color=logo.help_color, label=f'{logo.id}', linestyle = 'None', 
                                        markersize=5) )
            self.legend = self.ax.legend(handles=legend_elements,fontsize=self.group_id_size,
                            loc='upper right')
                            #, borderaxespad=0)
                            #,bbox_to_anchor=(1,1))
                            #,columnspacing=0,handletextpad=0,labelspacing=0,borderpad=0)
                            #bbox_transform=self.ax.get_figure().transFigure)
                            #bbox_to_anchor=(1, 1),
                            #bbox_transform=self.ax.get_figure().transFigure)
                            #loc='upper left',bbox_to_anchor=(1.005,1),borderaxespad=0.)
                            #plt.legend(bbox_to_anchor=(1, 1),
                            #bbox_transform=plt.gcf().transFigure)


    def draw_radiation_help(self):
        self.ax.add_patch(Circle(self.start_pos,self.radiation_radius,linewidth=1,fill=False,edgecolor='grey',alpha=0.5))
    
    def draw_hz_help(self):
        if not self.hide_x_ticks:
            maxlen = max([len(bits) for bits in self.seq_bits.values()]+[0])
            margin_ratio = 0
            if len(self.logos)>0:
                margin_ratio = self.logos[0].column_margin_ratio

            #self.ax.set_xticks([x+0.5+x*margin_ratio for x in range(maxlen)])
            self.ax.set_xticks([x+x*margin_ratio for x in range(maxlen+1)])
            self.ax.set_xticklabels([str(x) for x in range(maxlen+1)])
        
        if not self.hide_y_ticks:
            total_height = self.ceiling_pos[1]
            ticks = []
            ticklabels = []
            i = -1
            for logo in self.logos:
                i += 1
                start = logo.start_pos[1]
                height = logo.get_height()
                if i==0:
                    ticks.append(0)
                    ticklabels.append(0)
                ticks.append(start+height*(1+self.logo_margin_ratio))
                label = round(height*(1+self.logo_margin_ratio),2)
                if i == len(self.logos) - 1:
                    ticklabels.append('%s'%label)
                else:
                    ticklabels.append('%s/0'%label)
        
            self.ax.set_yticks(ticks)
            self.ax.set_yticklabels(ticklabels)



    def compute_positions(self):

        if self.logo_type == 'Radiation':
            self.start_pos = (0,0)
            self.radiation_radius = 1
            self.radiation_width = 1
            self.radiation_ratio = 0.9
            thetas = []
            head_ranges = [] 
            for group_id in self.group_ids:
                bit = self.seq_bits[group_id]
                head_range = max([sum([x[1] for x in col])*1.1  for col in bit][:self.radiation_head_n] + [0])
                head_ranges.append(head_range)
            self.radiation_radius = 2 * sum(head_ranges) / (np.pi*self.radiation_ratio)

            thetas = []
            for head_range in head_ranges:
                thetas.append(2*np.arctan(head_range/(2*self.radiation_radius)))

            deg_pointer = (np.pi/2 - sum(thetas))/2
            for index,theta in enumerate(thetas):
                c_deg = np.pi/2 - deg_pointer - theta/2
                deg_pointer += theta
                #start_pos = get_coor_by_angle(self.radiation_radius, c_deg, self.start_pos)
                logo = self.logos[index]
                logo.set_parent_start(self.start_pos)
                logo.set_start_pos((self.radiation_radius,0))
                logo.set_deg(c_deg)
                logo.set_radiation_space(head_ranges[index])
                logo.compute_positions()
        else:
            start_pos = self.start_pos
            if self.logo_type == 'Circle':
                start_pos = (self.start_pos[0], self.start_pos[1]+self.init_radius)
            elif self.logo_type == 'Threed':
                start_pos = (self.start_pos[0], self.start_pos[1], 0)
                self.start_pos = start_pos
            for index,logo in enumerate(self.logos):
                logo.set_parent_start(self.start_pos)
                logo.set_start_pos(start_pos)
                logo.compute_positions()
                if self.logo_type == 'Threed':
                    #start_pos = (start_pos[0], start_pos[1], start_pos[2] + self.threed_interval)
                    start_pos = (start_pos[0], start_pos[1], start_pos[2] + logo.get_height())
                else:
                    start_pos = (start_pos[0], start_pos[1] + logo.get_height()*(1+self.logo_margin_ratio))
            self.ceiling_pos = start_pos
    
    def get_height(self):
        return self.ceiling_pos[1] - self.start_pos[1]
    
    def get_max_logo_height(self):
        return max([logo.get_height() for logo in self.logos]+[0])

    def get_width(self):
        return max([logo.get_width() for logo in self.logos]+[0])

    def compute_xy(self):
        if self.logo_type == 'Horizontal':
            self.ax.set_xlim(self.start_pos[0]-1,self.start_pos[0] + self.get_width()+1)
            if self.ceiling_pos[1] > self.start_pos[1]:
                self.ax.set_ylim(self.start_pos[1],self.ceiling_pos[1])

            if self.show_group_id:
                r = self.ax.get_figure().canvas.get_renderer()
                x_range = 0
                for i in range(len(self.logos)):
                    text_width = self.logos[i].id_txt.get_window_extent(r).transformed(self.ax.transData.inverted()).width
                    logo_width = self.logos[i].get_width()
                    _range = text_width + 1 + logo_width
                    if _range > x_range:
                        x_range = _range
            
                if x_range > self.start_pos[0]:
                    self.ax.set_xlim(self.start_pos[0],x_range)


        elif self.logo_type == 'Circle':
            radius = self.ceiling_pos[1] - self.start_pos[1]
            if self.show_group_id:
                l = self.legend
                r = self.ax.get_figure().canvas.get_renderer()
                bbox = l.get_window_extent(r)
                bbox2 = bbox.transformed(self.ax.transAxes.inverted())
                padded = 2*radius/(1-bbox2.width) - 2*radius
                self.ax.set_xlim(self.start_pos[0] - radius,self.start_pos[0] + radius + padded)
                self.ax.set_ylim(self.start_pos[1] - radius,self.start_pos[1] + radius + padded)
            else:
                self.ax.set_xlim(self.start_pos[0] - radius,self.start_pos[0] + radius)
                self.ax.set_ylim(self.start_pos[1] - radius,self.start_pos[1] + radius)


        elif self.logo_type == 'Radiation':
            ###
            lims = []
            for logo in self.logos:
                width = logo.get_width()
                lim = max(width*np.sin(logo.deg), width*np.cos(logo.deg))
                lims.append(lim)
            self.ax.set_xlim(self.start_pos[0], self.start_pos[0] + self.radiation_radius + max(lims+[0]))
            self.ax.set_ylim(self.start_pos[1], self.start_pos[1] + self.radiation_radius + max(lims+[0]))

            if self.show_group_id:
                r = self.ax.get_figure().canvas.get_renderer()
                x_range = 0
                for i,logo in enumerate(self.logos):
                    lim = lims[i]
                    text_width = self.logos[i].id_txt.get_window_extent(r).transformed(self.ax.transData.inverted()).width
                    text_lim = max(text_width*np.sin(logo.deg), text_width*np.cos(logo.deg))
                    _range = text_width + text_lim + lim
                    if _range > x_range:
                        x_range = _range

                self.ax.set_xlim(self.start_pos[0], self.start_pos[0] + self.radiation_radius + x_range)
                self.ax.set_ylim(self.start_pos[0], self.start_pos[0] + self.radiation_radius + x_range)

        elif self.logo_type == 'Threed':
            self.ax.set_xlim(self.start_pos[0], self.start_pos[0] + self.get_width())
            self.ax.set_ylim(self.start_pos[1], self.start_pos[1] + sum([logo.get_height() for logo in self.logos[:-1]]) )
            self.ax.set_zlim(self.start_pos[1], self.start_pos[1] + sum([logo.get_height() for logo in self.logos]) )

    
    def set_figsize(self):
        if self.auto_size:
            if self.logo_type in ['Circle','Radiation','Threed']:
                self.ax.get_figure().set_figheight(10)
                self.ax.get_figure().set_figwidth(10)
            else:
                height =self.get_height()
                width = self.get_width()
                self.ax.get_figure().set_figheight(10)
                if height != 0:
                    self.ax.get_figure().set_figwidth(min(40,max(15,round((10/height)*width))))
        elif self.figure_size_x != -1 and self.figure_size_y != -1:
            self.ax.get_figure().set_figwidth(self.figure_size_x)
            self.ax.get_figure().set_figheight(self.figure_size_y)
        else:
            if self.logo_type == 'Horizontal':
                self.ax.get_figure().set_figheight(6)
                self.ax.get_figure().set_figwidth(12)
            else:
                self.ax.get_figure().set_figheight(10)
                self.ax.get_figure().set_figwidth(10)

    

    def get_entropy(self):
        ents = []
        #for grpid,probs in self.probs.items():

        seq_probs = self.probs
        if hasattr(self, 'full_probs'):
            seq_probs = self.full_probs
        
        for grpid in self.group_ids:
            probs = seq_probs[grpid]
            _ents = []
            _i = 0
            for pos in probs:
                _i += 1
                if len(pos) == 0:
                    _ents.append('-')
                    continue
                bases = set([x[0] for x in pos])
                if len(bases) == 1 and '-' in bases:
                    _ents.append('-')
                    continue
                _ep = [x[1] for x in pos if x[1]>0]
                entropy = -sum([x*np.log(x) for x in _ep ])
                _ents.append(entropy)
            if len(_ents) > 0:
                ents.append(_ents)
        
        return ents


    def get_group_mean_entropy_figure(self):

        if len(self.group_ids) == 0:
            return None
        fig,ax = plt.subplots()
        ents = self.get_entropy()
        mean_ents = [np.median([y for y in x if y!='-']) for x in ents]
        df = pd.DataFrame(mean_ents)
        df.index = self.group_ids
        df.columns = ['Mean Entropy of Positions']
        ax = df.plot.bar(ax=ax)
        ax.set_xlabel('Group')
        ax.set_ylabel('Mean Entropy of Positions')
        return ax
    
    def get_boxplot_entropy_figure(self):

        if len(self.group_ids) == 0:
            return None

        fig,ax = plt.subplots()
        ents = self.get_entropy()
        ents = [[y for y in x if y!='-'] for x in ents]

        lists = []
        for i in range(len(ents)):
            grp_id = self.group_ids[i]
            for item in ents[i]:
                lists.append([grp_id,item])
        if len(lists) == 0:
            return None
        df = pd.DataFrame(lists,columns=['Group','Entropy of Each Position'])
        return sns.boxplot(data=df,x='Group',y='Entropy of Each Position',ax=ax,order=self.group_ids)



    def get_entropy_figure(self):

        if len(self.group_ids) == 0:
            return None
        ents = self.get_entropy()[::-1]


        ent_df = pd.DataFrame(ents)
        ent_df = ent_df.replace('-',0)
        ent_df = ent_df.fillna(0)

        #ent_df = ent_df.loc[ent_df.index[::-1]]

        fig,ax = plt.subplots(figsize=(8,6))
        im = ax.imshow(ent_df)
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.1)

        cbar = ax.figure.colorbar(im, ax=ax,orientation='vertical',cax=cax)
        cbar.ax.set_xlabel('Entropy')

        kw = dict(horizontalalignment="center",
              verticalalignment="center",color='white')

        for i in range(len(ent_df)):
            for j in range(ent_df.columns.size):
                if j > len(ents[i]) - 1:
                    im.axes.text(j,i,'X',**kw)
                else:
                    if ents[i][j] == '-':
                        im.axes.text(j,i,'X',**kw)
                #else:
                #    im.axes.text(j,i,round(ents[i][j],2),**kw)


        ax.set_xticks(np.arange(ent_df.columns.size))
        ax.set_yticks(np.arange(len(self.group_ids)))
        ax.set_yticklabels(self.group_ids[::-1])
        ax.set_xticklabels(np.arange(ent_df.columns.size),rotation=90)
        ax.set_xlabel('Position')
        ax.set_ylabel('Group')
        ax.set_title('Entropy Heatmap')
        fig.tight_layout()

        ##extent = ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted())

        return fig
    
    def get_correlation(self):
        seq_bits = self.seq_bits        
        if hasattr(self, 'full_seq_bits'):
            seq_bits = self.full_seq_bits
        
        aligned_len =  len(seq_bits[self.group_ids[0]])
        if aligned_len < 1:
            return None
        

        bases = []
        for i in range(aligned_len):
            _bases = []
            for grp_id in self.group_ids:
                _bases += [x[0] for x in seq_bits[grp_id][i]]
            bases.append(sorted(list(set(_bases))))

        arrs = []
        for grp_id in self.group_ids:
            bits = seq_bits[grp_id]
            _arr = []
            for i in range(len(bits)):
                _dict = dict(bits[i])
                for base in bases[i]:
                    _arr.append(_dict.get(base,0))
            arrs.append(_arr)
        
        df = pd.DataFrame(arrs)
        df.index = self.group_ids
        return df.T.corr(method='pearson')
   

    def get_correlation_figure(self):

        if len(self.group_ids) == 0:
            return None
        fig,ax = plt.subplots()

        if (not self.padding_align) and (self.group_strategy != 'auto'):
            return None
        if len(self.group_ids) < 2:
            return None

        if hasattr(self,'correlation'):
            corr = self.correlation 
        else:
            corr = self.get_correlation()

        g = sns.clustermap(corr)
        g.ax_heatmap.tick_params(axis='both', which='major', labelsize=15)
        return g


    
    def get_grp_counts_figure(self):

        if len(self.group_ids) == 0:
            return None
        fig,ax = plt.subplots()
        lens = []
        if len(self.group_ids) == 0:
            return None
        for grp in self.group_ids:
            lens.append([grp,len(self.groups[grp])])
        df = pd.DataFrame(lens,columns=['Group','Counts'])
        ax = df.set_index('Group').plot.bar(ax=ax,y='Counts')
        for p in ax.patches:
            ax.annotate(str(p.get_height()), (p.get_x() + p.get_width()/2 , p.get_height() * 1.005),ha='center')
        ax.set_title('Sequence counts of each group')

        return ax
    
    def get_seq_lengths_dist(self):

        if len(self.seqs) == 0:
            return None

        fig,ax = plt.subplots()
        lens = [len(x[1]) for x in self.seqs]
        df = pd.DataFrame((pd.Series(lens).value_counts())).reset_index()
        df.columns = ['Length','Count']
        ax = df.set_index('Length').loc[sorted(set(lens))].plot.bar(ax=ax,y='Count')
        for p in ax.patches:
            ax.annotate(str(p.get_height()), (p.get_x() + p.get_width()/2 , p.get_height() * 1.005),ha='center')
        ax.set_title('Sequence lengths distribution')

        return ax
    

    #To do
    #def get_target_consist_score(self):
    #    if (self.target_group is not None) and (self.target_sequence_content is  not None):




#!/usr/bin/env python
import numpy as np

def compute_prob(groups,threshold=0):
    counts = {}
    for gid,group in groups.items():
        _counts = {}
        for name,seq in group:
            for i in range(len(seq)):
                if i not in _counts:
                    _counts[i] = {}
                #if seq[i] == '-':
                #    continue
                _counts[i][seq[i]] = _counts[i].get(seq[i],0) + 1
        counts[gid] = _counts
    
    probs = {}
    for gid,_counts in counts.items():
        _probs = []
        for i in range(len(_counts)):
            #total = sum(_counts[i].values())
            total = sum(_counts.get(i,{}).values())
            _ps = []
            #for base in sorted(_counts[i].keys()):
            for base in sorted(_counts.get(i,{}).keys()):
                ration = _counts[i][base]/total
                if ration < threshold:
                    continue
                _ps.append([base,ration])
            _probs.append(_ps)
        
        probs[gid] = _probs
    
    return probs
    
def compute_bits(groups, probs, seq_type='dna',no_correction=False):
    bits = {}
    for gid,prob in probs.items():
        if seq_type.lower() in ['dna','rna']:
            if no_correction:
                e = np.log2(4) 
            else:
                e = np.log2(4) - (4-1)/(np.log(2)*2*len(groups[gid]))
        elif seq_type.lower() == 'aa':
            if no_correction:
                e = np.log2(20)
            else:
                e = np.log2(20) - (20-1)/(np.log(2)*2*len(groups[gid]))
        bit = [] 
        for i in range(len(prob)):
            h = 0
            for base,p in prob[i]:
                #if base == '-':
                #    continue
                h -= p*np.log2(p) 
            _bit = [] 
            for base,p in prob[i]:
                if base == '-':
                    continue
                height = max(p * (e - h),0)
                _bit.append((base,height))
            bit.append(_bit)
        bits[gid] = bit 
    
    return bits
        
        


#!/usr/bin/env python
import sys
import os
import argparse

from .logo import LogoGroup
from .utils import read_file
from .colors import get_color_scheme
from .version import __version__
import matplotlib.pyplot as plt
import json
import uuid
import toml

def run_from_args(args):
    print('args: ', args)
    print('-----------------')
    print(f'uid: {args.uid}')

    os.makedirs(args.output_dir,exist_ok=True)
    os.makedirs(args.fa_output_dir,exist_ok=True)

    if args.color_scheme_json_file is not None:
        with open(args.color_scheme_json_file) as jsinf:
            color_scheme = json.load(jsinf)
    elif args.color_scheme_json_str is not None:
        color_scheme = json.loads(args.color_scheme_json_str)
    else:
        color_scheme = get_color_scheme(args.color_scheme)

    logogroup = LogoGroup(seqs=None, group_order = args.group_order, logo_type = args.type, group_strategy = args.group_strategy,
                          min_length = args.min_length, max_length = args.max_length, seq_file_type=args.seq_file_type,
                          align=args.align, align_metric=args.align_metric, connect_threshold = args.connect_threshold,
                          color=color_scheme, task_name=args.task_name, hide_left_axis = args.hide_left_axis,
                          hide_right_axis = args.hide_right_axis, hide_bottom_axis = args.hide_bottom_axis,
                          hide_top_axis = args.hide_top_axis, show_grid = args.show_grid, show_group_id = args.show_group_id,
                          hide_x_ticks = args.hide_x_ticks, hide_y_ticks = args.hide_y_ticks, hide_z_ticks=args.hide_z_ticks,
                          x_label=args.x_label, y_label=args.y_label, z_label=args.z_label,
                          title_size=args.title_size, label_size=args.label_size, group_id_size=args.group_id_size,
                          tick_size=args.tick_size, logo_margin_ratio = args.logo_margin_ratio, column_margin_ratio = args.column_margin_ratio,
                          figure_size_x=args.figure_size_x, figure_size_y=args.figure_size_y,
                          char_margin_ratio = args.char_margin_ratio, align_color=args.align_color,align_alpha=args.align_alpha ,
                          display_range_left=args.display_range_left, display_range_right=args.display_range_right,
                          gap_score = args.gap_score,
                          padding_align = args.padding_align,
                          hide_version_tag=args.hide_version_tag,
                          sequence_type = args.sequence_type,
                          height_algorithm=args.height_algorithm,
                          seq_file=args.seq_file, output_dir=args.output_dir,fa_output_dir=args.fa_output_dir,uid=args.uid,
                          group_resolution=args.group_resolution,clustering_method=args.clustering_method,
                          clustalo_bin=args.clustalo_bin,fasttreemp_bin=args.fasttreemp_bin,fasttree_bin=args.fasttree_bin,treecluster_bin=args.treecluster_bin,
                          withtree=args.withtree, group_limit = args.group_limit,
                          auto_size=args.auto_size
                          )
    if hasattr(logogroup,'error'):
        print('error:',logogroup.error)
        return {'error':logogroup.error}
    

    logogroup.draw()
    logogroup.savefig(f'{args.output_dir}/{args.output_name}')
    print(f'{args.output_dir}/{args.output_name}',' saved')
    base_name = '.'.join(args.output_name.split('.')[:-1]) 
    if not args.output_name.endswith('.png'):
        logogroup.savefig(f'{args.output_dir}/{base_name}.png')
        print(f'{args.output_dir}/{base_name}.png', ' saved')

    if args.analysis:

        fig = logogroup.get_grp_counts_figure()
        if fig:
            fig = fig.figure
            count_name = f'{args.output_dir}/{base_name}.counts.png'
            fig.savefig(count_name,bbox_inches='tight')
            plt.close(fig)

        fig = logogroup.get_seq_lengths_dist()
        if fig:
            fig = fig.figure
            lengths_name = f'{args.output_dir}/{base_name}.lengths.png'
            fig.savefig(lengths_name,bbox_inches='tight')
            plt.close(fig)


        fig = logogroup.get_entropy_figure()
        if fig:
            entropy_name = f'{args.output_dir}/{base_name}.entropy.png'
            fig.savefig(entropy_name,bbox_inches='tight')
            plt.close(fig)

        boxplot_entropy_name = f'{args.output_dir}/{base_name}.boxplot_entropy.png'
        fig = logogroup.get_boxplot_entropy_figure()
        if fig:
            fig = fig.figure
            fig.savefig(boxplot_entropy_name,bbox_inches='tight')
            plt.close(fig)

        if args.padding_align or args.group_strategy=='auto':
            clustermap_name = f'{args.output_dir}/{base_name}.clustermap.png'
            fig = logogroup.get_correlation_figure()
            if fig:
                fig.savefig(clustermap_name,bbox_inches='tight')
    
    return None
 

def run_from_config(config_file):

    config = toml.load(config_file)
    print(config)
    print('-----------------')

    uid = config['uid']
    print('uid: ', uid)

    os.makedirs(config['output_dir'],exist_ok=True)
    os.makedirs(config['fa_output_dir'],exist_ok=True)

    logogroup = LogoGroup(seqs=None, **config)

    if hasattr(logogroup,'error'):
        print('error:',logogroup.error)
        return {'error':logogroup.error}

    logogroup.draw()

    logogroup.savefig(f"{config['output_dir']}/{uid}.{config['logo_format']}")
    print(f"{config['output_dir']}/{uid}.{config['logo_format']}",' saved')

    if  config['logo_format'].lower() != 'png':
        logogroup.savefig(f"{config['output_dir']}/{uid}.png")
        print(f"{config['output_dir']}/{uid}.png', ' saved")
    
    

    if config['analysis']:
        format = 'png'
        if 'analysis_format' in config:
            format = config['analysis_format']

        fig = logogroup.get_grp_counts_figure()
        if fig:
            fig =fig.figure
            count_name = f"{config['output_dir']}/{uid}.counts.png"
            fig.savefig(count_name,bbox_inches='tight')
            if format != 'png':
                count_name = f"{config['output_dir']}/{uid}.counts.{format}"
                fig.savefig(count_name,bbox_inches='tight')
            plt.close(fig)

        fig = logogroup.get_seq_lengths_dist()
        if fig:
            fig = fig.figure
            lengths_name = f"{config['output_dir']}/{uid}.lengths.png"
            fig.savefig(lengths_name,bbox_inches='tight')
            if format != 'png':
                lengths_name = f"{config['output_dir']}/{uid}.lengths.{format}"
                fig.savefig(lengths_name,bbox_inches='tight')
            plt.close(fig)


        fig = logogroup.get_entropy_figure()
        if fig:
            entropy_name = f"{config['output_dir']}/{uid}.entropy.png"
            fig.savefig(entropy_name,bbox_inches='tight')
            if format != 'png':
                entropy_name = f"{config['output_dir']}/{uid}.entropy.{format}"
                fig.savefig(entropy_name,bbox_inches='tight')
            plt.close(fig)

        fig = logogroup.get_boxplot_entropy_figure()
        if fig:
            fig =fig.figure
            boxplot_entropy_name = f"{config['output_dir']}/{uid}.boxplot_entropy.png"
            fig.savefig(boxplot_entropy_name,bbox_inches='tight')
            if format != 'png':
                boxplot_entropy_name = f"{config['output_dir']}/{uid}.boxplot_entropy.{format}"
                fig.savefig(boxplot_entropy_name,bbox_inches='tight')

            plt.close(fig)

        if config.get('padding_align',False) or config.get('group_strategy','')=='auto':
            clustermap_name = f"{config['output_dir']}/{uid}.clustermap.png"
            fig = logogroup.get_correlation_figure()
            if fig:
                fig.savefig(clustermap_name,bbox_inches='tight')
                if format != 'png':
                    clustermap_name = f"{config['output_dir']}/{uid}.clustermap.{format}"
                    fig.savefig(clustermap_name,bbox_inches='tight')
    
    return None




def main():

    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    parser.add_argument('--config',type=str,help='The config file contain sequences',default=None)

    parser.add_argument('--type',type=str,help='Choose the layout type of sequence logo',choices=['Horizontal','Circle','Radiation','Threed'],default='Horizontal')
    parser.add_argument('--seq_file',type=str,help='The input file contain sequences')
    parser.add_argument('--seq_file_type',type=str,help='The type of input file', choices=['fasta','fastq'],default='fasta')
    parser.add_argument('--sequence_type',type=str,help='The type of sequences',choices=['auto','dna','rna','aa'],default='auto')

    #task
    parser.add_argument('--task_name',type=str,help='The title to displayed on the figure',default='MetaLogo')
    
    #sequences
    parser.add_argument('--min_length',type=int,help='The minimum length of sequences to be included',default=8)
    parser.add_argument('--max_length',type=int,help='The maximum length of sequences to be included',default=20)

    #group
    parser.add_argument('--group_strategy',type=str,help='The strategy to separate sequences into groups',choices=['auto','length','identifier'],default='auto')
    parser.add_argument('--clustering_method',type=str,help='The method for tree clustering',default='max')
    parser.add_argument('--group_resolution',type=float,help='The resolution for sequence grouping',default=0.5)
    parser.add_argument('--group_limit',type=int,help='The limit for group number',default=20)

    #sort
    parser.add_argument('--group_order',type=str,help='The order of groups',choices=['length','length_reverse','identifier','identifier_reverse'],default='length')

    #color
    parser.add_argument('--color_scheme',type=str,help='The color scheme',choices=['basic_dna_color','basic_rna_color','basic_aa_color'],default='basic_dna_color')
    parser.add_argument('--color_scheme_json_str',type=str,help='The json string of color scheme',default=None)
    parser.add_argument('--color_scheme_json_file',type=str,help='The json file of color scheme',default=None)

    #align
    parser.add_argument('--height_algorithm',type=str,help='The algorithm for character height',default='bits',choices=['bits','bits_without_correction','probabilities'])

    parser.add_argument('--align',action='store_true',dest='align', help='If show alignment of adjacent sequence logo')
    parser.add_argument('--padding_align',action='store_true',dest='padding_align', help='If padding logos to make multiple logo alignment')

    parser.add_argument('--align_metric',type=str,help='The metric for align score',default='dot_product',choices=['dot_product','js_divergence','cosine','entropy_bhattacharyya'])
    parser.add_argument('--connect_threshold',type=float,help='The align threshold',default=0.8)

    parser.add_argument('--gap_score',type=float,help='The gap score for alignment',default=-1.0)

    #display range
    parser.add_argument('--display_range_left',type=int,help='The start position of display range (Global alignment with padding required)',default=0)
    parser.add_argument('--display_range_right',type=int,help='Then end position of display range (Global alignment with padding requirement)',default=-1)

    #layout
    parser.add_argument('--withtree',action='store_true',dest='withtree', help='If show tree besides sequence logo')

    parser.add_argument('--logo_margin_ratio',type=float,help='Margin ratio between the logos',default=0.1)
    parser.add_argument('--column_margin_ratio',type=float,help='Margin ratio between the columns',default=0.05)
    parser.add_argument('--char_margin_ratio',type=float,help='Margin ratio between the chars',default=0.05)

    #style

    parser.add_argument('--hide_version_tag',action='store_true',dest='hide_version_tag',help='If show version tag of MetaLogo')

    parser.add_argument('--hide_left_axis',action='store_true',dest='hide_left_axis',help='If hide left axis')
    parser.add_argument('--hide_right_axis',action='store_true',dest='hide_right_axis',help='If hide right axis')
    parser.add_argument('--hide_top_axis',action='store_true',dest='hide_top_axis',help='If hide top axis')
    parser.add_argument('--hide_bottom_axis',action='store_true',dest='hide_bottom_axis',help='If hide bottom axis')

    parser.add_argument('--hide_x_ticks',action='store_true',dest='hide_x_ticks',help='If hide ticks of X axis')
    parser.add_argument('--hide_y_ticks',action='store_true',dest='hide_y_ticks',help='If hide ticks of Y axis')
    parser.add_argument('--hide_z_ticks',action='store_true',dest='hide_z_ticks',help='If hide ticks of Z axis')

    parser.add_argument('--x_label', type=str, help='The label for X axis')
    parser.add_argument('--y_label', type=str, help='The label for Y axis')
    parser.add_argument('--z_label', type=str, help='The label for Z axis')

    parser.add_argument('--show_group_id',action='store_true',dest='show_group_id',help='If show group ids')
    parser.add_argument('--show_grid',action='store_true',dest='show_grid',help='If show background grid')


    parser.add_argument('--title_size',type=int,help='The size of figure title',default=20)
    parser.add_argument('--label_size',type=int,help='The size of figure xy labels',default=10)
    parser.add_argument('--tick_size',type=int,help='The size of figure ticks',default=10)
    parser.add_argument('--group_id_size',type=int,help='The size of group labels',default=10)

    parser.add_argument('--figure_size_x',type=float,help='The width of figure',default=20)
    parser.add_argument('--figure_size_y',type=float,help='The height of figure',default=10)
    parser.add_argument('--auto_size',action='store_true',dest='auto_size',help='Let MetaLogo determine the size of figures')

    parser.add_argument('--align_color',type=str,help='The color of alignment',default='blue')
    parser.add_argument('--align_alpha',type=float,help='The transparency of alignment',default='0.2')

    #output 
    parser.add_argument('--output_dir',type=str,help='Output path of figure',default='figure_output')
    parser.add_argument('--output_name',type=str,help='Output name of figure',default='test.png')
    parser.add_argument('--fa_output_dir',type=str,help='Output path of fas',default='sequence_input')
    parser.add_argument('--uid',type=str,help='Task id',default=str(uuid.uuid4()))

    parser.add_argument('--logo_format',type=str,help='The format of figures',choices=['png','pdf'],default='png')

    #analysis
    parser.add_argument('--analysis',action='store_true',dest='analysis',help='If perform basic analysis on data')

    #software
    parser.add_argument('--clustalo_bin',type=str,help='The path of clustalo bin ',default='/usr/bin/clustalo')
    parser.add_argument('--fasttree_bin',type=str,help='The path of fasttree bin ',default='/usr/bin/FastTree')
    parser.add_argument('--fasttreemp_bin',type=str,help='The path of fasttreeMP bin ',default='/usr/bin/FastTreeMP')
    parser.add_argument('--treecluster_bin',type=str,help='The path of treecluster bin ',default='TreeCluster.py')

    parser.add_argument('-v', '--version', action='version', version=__version__)

    args = parser.parse_args()


    if args.config is not None:
        run_from_config(args.config)
    else:
        run_from_args(args)


if __name__ == '__main__':
    main()


#!/usr/bin/env python
from .character import Character
from .item import Item
from .utils import rotate
import numpy as np
from matplotlib.patches import PathPatch,Rectangle,Circle,Polygon
from matplotlib.path import Path
from .colors import get_color_scheme

basic_dna_color = get_color_scheme('basic_dna_color')


class Column(Item):

    def __init__(self, bases, weights, ax=None, start_pos=(0,0), logo_type = 'Horizontal', char_margin_ratio=0.05,
                 width=1, parent_start=(0,0), origin=(0,0), color=basic_dna_color, limited_char_width=None, 
                 path_dict={},*args, **kwargs):
        super(Column, self).__init__(*args, **kwargs)
        self.bases = bases 
        self.weights = weights 
        self.width = width
        self.char_margin_ratio = char_margin_ratio
        self.start_pos = start_pos
        self.parent_start = parent_start
        self.origin = origin
        self.logo_type = logo_type
        self.color = color
        self.limited_char_width = limited_char_width
        self.path_dict = path_dict

        #self.path_hight, self.init_hight, self.target_height = self.get_heights()
        self.characters = []
        if ax == None:
            self.generate_ax(threed=(self.logo_type=='Threed'))
        else:
            self.ax = ax

        if limited_char_width == None:
            self.limited_char_width = self.get_limited_char_width()

        self.generate_components()

    def generate_components(self):
        for base,weight in sorted(zip(self.bases,self.weights),key=lambda d:d[1]):
            character = Character(base,width=self.width,height=weight,ax=self.ax,
                                    logo_type=self.logo_type, parent_start=self.start_pos,
                                    origin=self.origin,color=self.color,limited_char_width=self.limited_char_width,
                                    path_dict=self.path_dict)
            self.characters.append(character)
    
    def draw(self):
        for character in self.characters:
            character.draw()
        
    def draw_wrap(self):
        p1,p2,p3,p4 = self.get_edge()
        verts = [p1,p2,p3,p4,p1]
        codes = [
            Path.MOVETO,
            Path.LINETO,
            Path.LINETO,
            Path.LINETO,
            Path.CLOSEPOLY
        ]
        self.ax.add_patch(PathPatch(Path(verts, codes)))
    
    def compute_positions(self):
        start_pos = self.start_pos
        for character in self.characters:
            character.set_start_pos(start_pos)
            character.set_parent_start(self.start_pos)

            if self.logo_type == 'Circle':
                character.set_deg(self.deg)
            elif self.logo_type == 'Radiation':
                character.set_deg(self.deg)
                character.set_radiation_space(self.radiation_space)

            character.set_width(self.width)
            character.compute_positions()
            if self.logo_type == 'Threed':
                start_pos = (start_pos[0], start_pos[1] + character.get_height() *(1+self.char_margin_ratio), start_pos[2])
            else:
                start_pos = (start_pos[0], start_pos[1] + character.get_height() * (1+self.char_margin_ratio))
    
    def get_height(self):
        height = sum([char.get_height() * (1+self.char_margin_ratio) for char in self.characters[:-1]]) 
        if len(self.characters) > 0:
            height += self.characters[-1].get_height()
        return height

    def get_width(self):
        return max([char.get_width()  for char in self.characters]+[0])
    
    def get_edge(self):

        h = self.get_height()
        w = self.get_width()

        if self.logo_type in 'Horizontal':
            leftbottom = self.start_pos
            rightbottom = (self.start_pos[0]+w, self.start_pos[1])
            righttop = (self.start_pos[0] + w, self.start_pos[1] + h ) 
            lefttop = (self.start_pos[0],self.start_pos[1] + h)
            return leftbottom,rightbottom,righttop,lefttop

        if self.logo_type in 'Threed':
            leftbottom = self.start_pos
            rightbottom = (self.start_pos[0]+w, self.start_pos[1], self.start_pos[2])
            righttop = (self.start_pos[0] + w, self.start_pos[1] + h, self.start_pos[2]) 
            lefttop = (self.start_pos[0],self.start_pos[1] + h, self.start_pos[2])
            return leftbottom,rightbottom,righttop,lefttop

        if self.logo_type == 'Circle':
            p1 = (self.start_pos[0] - w/2, self.start_pos[1])
            p2 = (self.start_pos[0] + w/2, self.start_pos[1])
            p3 = (self.start_pos[0] + w/2, self.start_pos[1]+h)
            p4 = (self.start_pos[0] - w/2, self.start_pos[1]+h)
            nodes = rotate([p1,p2,p3,p4],origin=self.start_pos, angle=self.deg-np.pi/2)
            return nodes
       
        if self.logo_type == 'Radiation':
            p1 = (self.start_pos[0], self.start_pos[1]-self.radiation_space/2)
            p2 = (self.start_pos[0]+w, self.start_pos[1]-self.radiation_space/2)
            p3 = (self.start_pos[0]+w, self.start_pos[1]+h-self.radiation_space/2)
            p4 = (self.start_pos[0], self.start_pos[1]+h-self.radiation_space/2)
            nodes = rotate([p1,p2,p3,p4],origin=self.origin, angle=self.deg)
            return nodes
        


 



        




#!/usr/bin/env python

basic_dna_color_scheme = {'A': '#009980',
        'C': '#59B3E6',
        'G': '#E69B04',
        'T': '#1A1A1A',
        '-': 'grey'}

basic_rna_color_scheme = {'A': '#009980',
        'C': '#59B3E6',
        'G': '#E69B04',
        'U': '#1A1A1A',
        '-': 'grey'}


basic_aa_color_scheme ={ #https://jbloomlab.github.io/dmslogo/dmslogo.colorschemes.html
        'A': 'black',
        'C': 'green',
        'D': 'red',
        'E': 'red',
        'F': 'black',
        'G': 'green',
        'H': 'blue',
        'I': 'black',
        'K': 'blue',
        'L': 'black',
        'M': 'black',
        'N': '#FF00AE',
        'P': 'black',
        'Q': '#FF00AE',
        'R': 'blue',
        'S': 'green',
        'T': 'green',
        'V': 'black',
        'W': 'black',
        'Y': 'green'
    }
 
def get_color_scheme(scheme):
    if scheme == 'basic_dna_color':
        return basic_dna_color_scheme
    if scheme == 'basic_rna_color':
        return basic_rna_color_scheme
    if scheme == 'basic_aa_color':
        return basic_aa_color_scheme
    return None



#!/usr/bin/env python

from matplotlib import pyplot as plt
from matplotlib.textpath import TextPath

class Item():
    def __init__(self, *args, **kwargs) :
        pass

    def generate_ax(self,threed=False,withtree=False):
        if threed:
            fig, ax = plt.subplots(1, 1,figsize=(10,10),subplot_kw=dict(projection="3d"))
            self.ax = ax    
        elif withtree: 
            fig, (ax1,ax2) = plt.subplots(1, 2,figsize=(10,10),gridspec_kw={'width_ratios': [1, 4]})
            self.ax = ax2
            self.ax0 = ax1
            plt.subplots_adjust(wspace=0)
        else:
            fig, ax = plt.subplots(1, 1,figsize=(10,10))
            self.ax = ax    
    
    def draw(self):
        pass
    
    def savefig(self,filename,bbox_inches='tight'):
        if (hasattr(self,'ax')) and (self.ax is not None):
            self.ax.get_figure().savefig(filename,bbox_inches=bbox_inches)

    def set_start_pos(self,start_pos):
        self.start_pos = start_pos
    
    def set_parent_start(self,parent_start):
        self.parent_start = parent_start 

    def set_deg(self,deg):
        self.deg = deg
    
    def set_width(self,width):
        self.width = width

    def set_radiation_space(self,space):
        self.radiation_space = space

    def get_limited_char_width(self, limited_char='E'):
        tmp_path = TextPath((0, 0), 'E', size=1)
        return tmp_path.get_extents().width



#!/usr/bin/env python
from numpy.core.fromnumeric import product
from scipy.stats import spearmanr,pearsonr
import numpy as np
from scipy.spatial import distance
import math

def dotproduct(v1, v2):
    return sum((a*b) for a, b in zip(v1, v2))
def length(v):
    return math.sqrt(dotproduct(v, v))
def costheta(v1, v2):
    return dotproduct(v1, v2) / (length(v1) * length(v2))

def get_score_mat(bits_array, align_metric = 'sort_consistency',  gap_score=-1, seq_type='dna'):
    scores_mat = {}
    for i in range(len(bits_array)):
        for j in range(len(bits_array)):
            if i >= j:
                continue
            bits1 = bits_array[i]
            bits2 = bits_array[j]
            align1,align2 = needle(bits1,bits2, align_metric=align_metric, 
                                    gap_penalty=gap_score, seq_type=seq_type)
            score = 0
            for pos1, pos2 in zip(align1,align2):
                if pos1 == '-' or pos2 == '-':
                    score += gap_score
                else:
                    score += match_score(bits1[pos1],bits2[pos2], align_metric=align_metric,seq_type=seq_type)

            if i not in scores_mat:
                scores_mat[i] = {}
            scores_mat[i][j] = score/len(align1)
    return scores_mat

def msa(bits_array, scores_mat, align_metric = 'sort_consistency', gap_score=-1, seq_type='dna'):

    #find the nearest couple
    max_score = max([max(scores_mat[x].values()) for x in scores_mat])
    findij = False
    for i in scores_mat:
        for j in scores_mat[i]:
            #if abs(scores_mat[i][j] - max_score) < 0.00001:
            if scores_mat[i][j] == max_score:
                findij = True
                break
        if findij:
            break
    #align the first two
    align1,align2 = needle(bits_array[i],bits_array[j], align_metric=align_metric, 
                              gap_penalty=gap_score, seq_type=seq_type)
    #print(align1)#test
    #print(align2)#test
    
    pools = [i,j]
    new_bits_array = []
    new_bits_array.append([bits_array[i][pos] if pos!= '-' else [] for pos in align1])
    new_bits_array.append([bits_array[j][pos] if pos!= '-' else [] for pos in align2])
    repeat = 0
    while len(pools) < len(bits_array):
        repeat += 1
        if repeat > len(bits_array) + 1:
            break
        left = set(range(len(bits_array))) - set(pools)
        max_score = -1E9
        max_i= -1
        max_j = -1

        for i in pools:
            for j in left:
                score = scores_mat[min(i,j)][max(i,j)]
                if score > max_score:
                    max_score = score
                    max_i = i
                    max_j = j
        #
        bits1 = new_bits_array[pools.index(max_i)]
        bits2 = bits_array[max_j]
        align1,align2 = needle(bits1,bits2, align_metric=align_metric, 
                              gap_penalty=gap_score, seq_type=seq_type)
        
        for i in range(len(new_bits_array)):
            _arr = []
            for pos in align1:
                if pos == '-':
                    _arr.append([])
                else:
                    _arr.append(new_bits_array[i][pos])
            new_bits_array[i] = _arr

        new_bits_array.append([bits2[pos] if pos!= '-' else [] for pos in align2]) 
        pools.append(max_j)
    
    sorted_bits_array = []
    for i in range(len(pools)):
        sorted_bits_array.append(new_bits_array[pools.index(i)])
    
    return sorted_bits_array



def get_connect(bits_array, align_metric = 'sort_consistency', gap_score=-1, msa_input=False, seq_type='dna'):
    connected = {}
    for index,bit in enumerate(bits_array):
        if index == len(bits_array) - 1:
            break
        bits1 = bit
        bits2 = bits_array[index + 1]
        if msa_input:
            align1 = list(range(len(bits1)))
            align2 = list(range(len(bits2)))
        else:
            align1,align2 = needle(bits1,bits2, align_metric=align_metric, 
                                gap_penalty=gap_score,seq_type=seq_type)
        connected[index] = {}

        for pos1, pos2 in zip(align1,align2):
            if pos1 == '-' or pos2 == '-':
                continue
            score = match_score(bits1[pos1],bits2[pos2], align_metric=align_metric, seq_type = seq_type)
            connected[index][pos1] = [score, [pos2]]
    return connected

max_entropy_aa = -sum([(1/20)*np.log(1/20) for i in range(20)])
max_entropy_dna = -sum([(1/4)*np.log(1/4) for i in range(4)])

def match_score(bit1, bit2, align_metric='sort_consistency',gap_score=-1,seq_type='dna'):

    try:

        if len(bit1) == 0 or len(bit2) == 0:
            return 0

        if align_metric not in ['dot_product','sort_consistency','js_divergence','cosine','entropy_bhattacharyya']:
            align_metric = 'dot_product'
        
        if align_metric == 'entropy_bhattacharyya':
            bit1 = dict(bit1)
            bit2 = dict(bit2)
            keys = sorted(list(bit1.keys()|bit2.keys()))
            v1 = [bit1.get(key,0) for key in keys]
            v2 = [bit2.get(key,0) for key in keys]
            bc = sum([np.sqrt(i1*i2) for i1,i2 in zip(v1,v2)])
            max_entropy = 0
            if seq_type.lower() in  ['protein','aa']:
                max_entropy = max_entropy_aa
            if seq_type.lower() in  ['dna','rna']:
                max_entropy = max_entropy_dna
            entropy1 = -sum([bit1.get(key,0)*np.log(bit1.get(key,0)) for key in keys if bit1.get(key,0) > 0])
            entropy2 = -sum([bit2.get(key,0)*np.log(bit2.get(key,0)) for key in keys if bit2.get(key,0) > 0])
            res = bc * np.sqrt((1 - (entropy1/max_entropy)) * (1 - (entropy2/max_entropy)))

            return res


        if align_metric == 'dot_product':
            bit1 = dict(bit1)
            bit2 = dict(bit2)
            keys = sorted(list(bit1.keys()|bit2.keys()))
            v1 = [bit1.get(key,0) for key in keys]
            v2 = [bit2.get(key,0) for key in keys]
            val = dotproduct(v1,v2)
            return val

        if align_metric == 'cosine':
            bit1 = dict(bit1)
            bit2 = dict(bit2)
            keys = sorted(list(bit1.keys()|bit2.keys()))
            v1 = [bit1.get(key,0) for key in keys]
            v2 = [bit2.get(key,0) for key in keys]
            if length(v1)*length(v2)==0:
                return 0
            return costheta(v1,v2)

        if align_metric == 'sort_consistency':
            bit1 = sorted(bit1, key=lambda d:d[1],reverse=True)
            bit2 = sorted(bit2, key=lambda d:d[1],reverse=True)
            score = 0
            for i in range(min(len(bit1),len(bit2))):
                if bit1[i][0] == bit2[i][0]:
                    score += bit1[i][1] * bit2[i][1]
            return score
    
        if align_metric =='js_divergence': #noted, here must input probabilites rather than bits.
            q1 = []
            q2 = []
            bit1 = dict(bit1)
            bit2 = dict(bit2)
            keys = sorted(list(bit1.keys()|bit2.keys()))
            for key in keys:
                q1.append(bit1.get(key,0))
                q2.append(bit2.get(key,0))
            if sum(q1)*sum(q2) == 0:
                return 0
            return 1-distance.jensenshannon(q1,q2)

    except Exception as e:
        print('exception: ', e)
        return 0


#https://github.com/alevchuk/pairwise-alignment-in-python/blob/master/alignment.py
def needle(seq1, seq2, gap_penalty=-1, align_metric='sort_consistency',seq_type='dna'):
    m, n = len(seq1), len(seq2)  # length of two sequences
    
    # Generate DP table and traceback path pointer matrix
    score = np.zeros((m+1, n+1))      # the DP table
   
    # Calculate DP table
    for i in range(0, m + 1):
        score[i][0] = gap_penalty * i
    for j in range(0, n + 1):
        score[0][j] = gap_penalty * j
    for i in range(1, m + 1):
        for j in range(1, n + 1):
            match = score[i - 1][j - 1] + match_score(seq1[i-1], seq2[j-1],align_metric=align_metric,seq_type=seq_type)
            delete = score[i - 1][j] + gap_penalty
            insert = score[i][j - 1] + gap_penalty
            score[i][j] = max(match, delete, insert)


    # Traceback and compute the alignment 
    align1, align2 = [], [] 
    i,j = m,n # start from the bottom right cell
    repeat = 0
    while i > 0 and j > 0: # end toching the top or the left edge
        repeat += 1
        if repeat > (m+1) * (n*1):
            break 
        score_current = score[i][j]
        score_diagonal = score[i-1][j-1]
        score_up = score[i][j-1]
        score_left = score[i-1][j]

        #print('seq1[i-1]:', seq1[i-1])
        #print('seq2[j-1]:', seq1[j-1])

        if score_current == score_diagonal + match_score(seq1[i-1], seq2[j-1],align_metric=align_metric,seq_type=seq_type):
            align1.append(i-1)
            align2.append(j-1)
            i -= 1
            j -= 1
        elif score_current == score_left + gap_penalty:
            align1.append(i-1)
            align2.append('-')
            i -= 1
        elif score_current == score_up + gap_penalty:
            align1.append('-')
            align2.append(j-1)
            j -= 1
        else:
            break

    # Finish tracing up to the top left cell
    while i > 0:
        align1.append(i-1)
        align2.append('-')
        i -= 1
    while j > 0:
        align1.append('-')
        align2.append(j-1)
        j -= 1
    #print('align1:',  align1)
    #print('align2: ', align2)
    return align1[::-1],align2[::-1]



#!/usr/bin/env python
from collections import namedtuple
import os
import uuid
import pandas as pd
import re
import dendropy
import matplotlib
import matplotlib.pyplot as plt
import seaborn as sns
from Bio import Phylo
from ete3 import Tree

def get_distance_range(tree_file):
    tree = dendropy.Tree.get(path=tree_file,schema='newick')
    pdc = tree.phylogenetic_distance_matrix()
    dists = pdc.distances()
    return dists

def get_distance_range_lessmem(tree_file):
    tree = Tree(tree_file)
    dists = []
    for node in tree:
        dists.append(node.get_farthest_node()[1])
    return dists


def get_score_df(score_f):
    arrs = []
    with open(score_f,'r') as inpf:
        for line in inpf:
            line = line.strip()
            if len(line) == 0:
                continue
            if line[0] == '#':
                continue
            _arrs = re.split(' +',line)
            arrs.append(_arrs[:3])
    if len(arrs) == 0:
        return None
    df = pd.DataFrame(arrs,columns = ['POS','SEQ','SCORE'])
    df['SCORE'] = df['SCORE'].map(float)
    df['BASE'] = df['POS'] + '-' + df['SEQ']
    return df

def drawdists(dists,output):
    fig,ax = plt.subplots()
    df = pd.DataFrame({'Pairwise distance':list(dists)})
    g = sns.histplot(df,x='Pairwise distance',kde=True,ax=ax)
    fig.savefig(output,bbox_inches='tight')
    return

def drawtree(input,output):
    tree = Phylo.read(input, 'newick')
    tree.ladderize()  # Flip branches so deeper clades are displayed at top

    matplotlib.rc('font', size=10)
    fig = plt.figure(figsize=(20, 20), dpi=100)
    axes = fig.add_subplot(1, 1, 1)
    Phylo.draw(tree, axes=axes)
    plt.savefig(output, dpi=100)


def reverse_msa_seqname(name_dict,oldfile,newfile):

    with open(newfile,'w') as outpf:
        seqname = ''
        seq = ''
        with open(oldfile,'r') as inpf:
            for line in inpf:
                line = line.strip()
                if len(line) == 0:
                    continue
                if line[0] == '>':
                    if seqname != '' and seq != '':
                        for alias in name_dict.get(seqname,[]):
                            outpf.write(f'>{alias}\n')
                            outpf.write(f'{seq}\n')
                    seqname = line[1:]
                    seq = ''
                else:
                    seq += line

            if seqname != '' and seq != '':
               for alias in name_dict.get(seqname,[]):
                   outpf.write(f'>{alias}\n')
                   outpf.write(f'{seq}\n')
    return None

def reverse_tree_seqname(name_dict,oldtreefile,newtreefile):
    tree = dendropy.Tree.get(path=oldtreefile,schema='newick')
    for node in tree:
        if node.is_leaf():
            if node.taxon is not None:
                if len(name_dict.get(node.taxon.label,[])) > 0 :
                    node.taxon.label = name_dict.get(node.taxon.label,[])[0]
    tree.write(path=newtreefile,schema='newick')
    return None

def save_group_seqs(group_dict,outfa):
    with open(outfa,'w') as outpf:
        outpf.write('Sequence name\tGroup_id\n')
        for grpid in group_dict:
            for seqname,seq in group_dict[grpid]:
                outpf.write(f'{seqname} \t {grpid} \n')

def save_seqs(seqs, filename):
    with open(filename,'w') as outpf:
        for seqname,seq in seqs:
            outpf.write(f'>{seqname}\n')
            outpf.write(f'{seq}\n') 

def auto_detect_groups(seqs, seq_fa, sequence_type='aa',group_resolution=1,clustering_method='max', 
                       clustalo_bin='',fasttree_bin='',fasttreemp_bin='',treecluster_bin='',
                       uid='', fa_output_dir='', figure_output_dir=''):
    
    if seq_fa == '':
        if uid == '':
            uid = str(uuid.uuid4())
        seq_fa = f'{fa_output_dir}/server.{uid}.fa'

    if not os.path.exists(seq_fa): 
        save_seqs(seqs, seq_fa)

    dep_seq_fa = f'{fa_output_dir}/server.{uid}.dep.fa'
    name_dict, seq_dict = deduplicate(seq_fa,dep_seq_fa)

    msa_dict = msa(dep_seq_fa,f'{fa_output_dir}/server.{uid}.msa.fa',clustalo_bin)

    if len(seqs) > 1000:
        fasttree(f'{fa_output_dir}/server.{uid}.msa.fa',
                f'{fa_output_dir}/server.{uid}.fasttree.tree',
                fasttreemp_bin,sequence_type)
    else:
        fasttree(f'{fa_output_dir}/server.{uid}.msa.fa',
                f'{fa_output_dir}/server.{uid}.fasttree.tree',
                fasttree_bin,sequence_type)
    try:
        if os.path.exists(f'{fa_output_dir}/server.{uid}.treedists.csv'):
            dists = pd.read_csv(f'{fa_output_dir}/server.{uid}.treedists.csv',index_col=False,header=0)['0'].tolist()
        else: 
            if len(seqs) > 1000:
                dists = get_distance_range_lessmem(f'{fa_output_dir}/server.{uid}.fasttree.tree')
            else:
                dists = get_distance_range(f'{fa_output_dir}/server.{uid}.fasttree.tree')
            pd.Series(list(dists)).to_csv(f'{fa_output_dir}/server.{uid}.treedists.csv',index=None)
    except:
        dists = get_distance_range_lessmem(f'{fa_output_dir}/server.{uid}.fasttree.tree')

    
    treecluster(group_resolution,clustering_method,dists,f'{fa_output_dir}/server.{uid}.fasttree.tree',f'{fa_output_dir}/server.{uid}.fasttree.cluster',treecluster_bin)

    reverse_msa_seqname(name_dict,f'{fa_output_dir}/server.{uid}.msa.fa',f'{fa_output_dir}/server.{uid}.msa.rawid.fa')
    reverse_tree_seqname(name_dict,f'{fa_output_dir}/server.{uid}.fasttree.tree',f'{fa_output_dir}/server.{uid}.fasttree.rawid.tree')

    drawdists(dists,f'{figure_output_dir}/{uid}.treedistances.png')
    drawtree(f'{fa_output_dir}/server.{uid}.fasttree.rawid.tree',f'{figure_output_dir}/{uid}.tree.png')

    cluster_df = pd.read_csv(f'{fa_output_dir}/server.{uid}.fasttree.cluster',sep='\t')
    groups_dict = {}
    for index, grp in cluster_df.groupby('ClusterNumber'):
        #if str(index) == '-1':
        #    continue
        groups_dict[index] = []
        for seqname in grp['SequenceName']:
            for _seqname in name_dict[seqname]:
                groups_dict[index].append([_seqname,msa_dict[seqname]])
    save_group_seqs(groups_dict,f'{fa_output_dir}/server.{uid}.grouping.fa')
        
    return groups_dict

def msa(seq_fa,outfile,clustalo_bin):
    if not os.path.exists(outfile):
        cmd = f'{clustalo_bin} --auto -i {seq_fa} -o {outfile}'
        os.system(f'{clustalo_bin} --auto -i {seq_fa} -o {outfile}')
    msa_dict = {}
    with open(outfile,'r') as inpf:
        seqname = ''
        seq = ''
        for line in inpf:
            line = line.strip()
            if line[0] == '>':
                if seq != '':
                    msa_dict[seqname] = seq
                seqname = line[1:]
                seq = ''
            else:
                seq += line
        if seq != '':
            msa_dict[seqname] = seq
    return msa_dict

def fasttree(msa_fa,outfile_tree,fasttree_bin='',sequence_type='aa'):
    if (not os.path.exists(outfile_tree)):
        if sequence_type != 'aa':
            cmd = f'{fasttree_bin} -nt  -quiet -nopr {msa_fa} > {outfile_tree} '
        else:
            cmd = f'{fasttree_bin}  -quiet -nopr {msa_fa} > {outfile_tree} '
        return os.system(cmd)
    return -1

def treecluster(threshold,clustering_method,dists,treefile,outfile,treecluster_bin=''):
    if threshold == 0:
        adj_threshold = threshold
    else:
        sorted_dists = sorted(dists)
        adj_threshold_idx = round(threshold*len(dists))
        adj_threshold = sorted_dists[min(len(dists)-1,adj_threshold_idx)]
    cmd = f'{treecluster_bin} -i {treefile} -o {outfile} -t {adj_threshold} -m {clustering_method}'
    return os.system(cmd)

def deduplicate(seq_fa,out_fa):
    seq_dict = {}
    name_dict = {}
    seq2seqno = {}
    with open(out_fa,'w') as outpf:
        with open(seq_fa,'r') as inpf:
            seq = ''
            seqname = ''
            seq_no = 0
            for line in inpf:
                line = line.strip()
                if line[0]=='>':
                    if seq != '' and seqname != '':
                        if seq not in seq2seqno:
                            seq_no += 1
                            seq_dict[f'seq-{seq_no}'] = seq
                            name_dict[f'seq-{seq_no}'] = [seqname]
                            outpf.write(f'>seq-{seq_no}\n')
                            outpf.write(f'{seq}\n')
                            seq2seqno[seq] = f'seq-{seq_no}'
                        else:
                            name_dict[f'seq-{seq_no}'].append(seqname)

                    seq = ''
                    seqname = line[1:]
                else:
                    seq += line.upper()
            if seq != '' and seqname != '':
                if seq not in seq2seqno:
                    seq_no += 1
                    seq_dict[f'seq-{seq_no}'] = seq
                    name_dict[f'seq-{seq_no}'] = [seqname]
                    outpf.write(f'>seq-{seq_no}\n')
                    outpf.write(f'{seq}\n')
                    seq2seqno[seq] = f'seq-{seq_no}'
                else:
                    name_dict[f'seq-{seq_no}'].append(seqname)
    del seq2seqno 
    return name_dict, seq_dict





#!/usr/bin/env python

from matplotlib import pyplot as plt
from matplotlib import patches
import numpy as np
import math
from matplotlib.transforms import Affine2D, Bbox
from matplotlib.textpath import TextPath
from matplotlib.patches import PathPatch,Rectangle,Circle,Polygon
from matplotlib.path import Path
import mpl_toolkits.mplot3d.art3d as art3d


from .item import Item
from .colors import get_color_scheme
from .utils import rotate

basic_dna_color = get_color_scheme('basic_dna_color')

class Character(Item):

    def __init__(self, char, ax=None, start_pos=(0,0),  width=1, height=1, limited_char_width=None, 
                    logo_type='Horizontal', font = 'Arial', color = basic_dna_color, alpha = 1, 
                    parent_start=(0,0), deg=np.pi/2, origin=(0,0), path_dict={},*args, **kwargs):
        super(Character, self).__init__(*args, **kwargs)
        self.char = char
        self.start_pos = start_pos
        self.width = width 
        self.height =  height
        self.logo_type = logo_type
        self.alpha = alpha
        self.parent_start = parent_start
        self.origin = origin
        self.deg = deg
        self.path = None
        self.patch = None
        self.color_map = color
        self.limited_char_width = limited_char_width
        self.path_dict = path_dict

        if ax == None:
            self.generate_ax(threed=(self.logo_type=='Threed'))
        else:
            self.ax = ax
        if limited_char_width == None:
            self.limited_char_width = self.get_limited_char_width()
        self.generate_components()
    
    def generate_components(self):
        self.path = TextPath(self.start_pos, self.char, size=1)
    
    def transform_path(self, transformation):
        return transformation.transform_path(self.path)


    def set_font(self, font):
        self.font = font

    def set_alpha(self, alpha):
        self.set_alpha = alpha
    
    def get_path_extents(self):
        return self.path.get_extents()

    def get_patch_extents(self):
        return self.patch.get_extents()
    
    def transform(self):
        width = self.width
        height = self.height

        if self.char in self.path_dict:
            tmp_path,bbox = self.path_dict[self.char]
        else:
            tmp_path = TextPath((0,0), self.char, size=1)
            bbox = tmp_path.get_extents()

        if self.logo_type in ['Horizontal','Threed']:
            hoffset = (width - bbox.width * width / max(bbox.width,self.limited_char_width))/2
            voffset = 0
        elif self.logo_type == 'Circle':
            hoffset = -1*(bbox.width * width / max(bbox.width,self.limited_char_width))/2
            voffset = 0
        elif self.logo_type == 'Radiation':
            hoffset = 0 
            voffset = -1 * self.radiation_space/2
        else:
            pass

        transformation = Affine2D() \
            .translate(tx=-bbox.xmin, ty=-bbox.ymin) \
            .scale(sx=width/max(bbox.width,self.limited_char_width), sy=height/bbox.height) \
            .translate(tx=self.start_pos[0] + hoffset,ty=self.start_pos[1] + voffset)
        
        if self.logo_type == 'Circle':
            transformation = transformation.rotate_around(self.parent_start[0], self.parent_start[1], self.deg-np.pi/2) 
        elif self.logo_type == 'Radiation':
            transformation = transformation.rotate_around(self.origin[0], self.origin[1], self.deg)
            #pass
        
        #print('self.parent_start: ',self.parent_start)
        
        #self.ax.annotate(self.start_pos, self.deg)

        self.path = transformation.transform_path(tmp_path)
        self.patch = PathPatch(self.path, linewidth=0, 
                                facecolor=self.color_map.get(self.char,self.color_map.get('other','grey')),                              
                                alpha=self.alpha,
                                edgecolor=self.color_map.get(self.char,self.color_map.get('other','grey')))
        

    def draw(self):
        self.transform()
        self.ax.add_patch(self.patch)
        if self.logo_type == 'Threed':
            art3d.pathpatch_2d_to_3d(self.patch, z=self.start_pos[2], zdir='y')

    def compute_positions(self):
        pass

    def get_height(self):
        return self.height

    def get_width(self):
        return self.width



#!/usr/bin/env python

try: 
    from setuptools import setup
except ImportError: 
    from distutils.core import setup

exec(open('MetaLogo/version.py').read())

setup(name='MetaLogo',
      version=__version__,
      description='MetaLogo is a heterogeneity-aware sequence logo generator and aligner',
      long_description='MetaLogo is a tool for making sequence logos. It can take multiple sequences as input, automatically identify the homogeneity and heterogeneity among sequences and cluster them into different groups given any wanted resolution, finally output multiple aligned sequence logos in one figure. Grouping can also be specified by users, such as grouping by lengths, grouping by sample Id, etc.  Compared to conventional sequence logo generator, MetaLogo can display the total sequence population in a more detailed, dynamic and informative view. homogeneity.\
                        To use MetaLogo, you could visit our public webserver http://metalogo.omicsnet.org. You could also install MetaLogo as a python package to using MetaLogo in your python scripts or in your OS terminal. If you want to provide MetaLogo to people in your local network, you could also setup a webserver by using docker.\
                        Please check the tutorial for detailed usage of MetaLogo package and webserver (https://github.com/labomics/MetaLogo).',
      long_description_content_type = 'text/plain',
      author='Yaowen Chen',
      author_email='achenge07@163.com',
      url='https://github.com/labomics/MetaLogo',
      packages=['MetaLogo'],
      entry_points={
          'console_scripts': ['metalogo=MetaLogo.entry:main']
      },
      python_requires='>=3.6',                
      install_requires=[
          'biopython>=1.77',
          'matplotlib>=3.3.0',
          'numpy>=1.19.1',
          'pandas>=1.3.0',
          'scipy>=1.5.2',
          'seaborn>=0.11.1',
          'toml>=0.10.2',
          'treecluster>=1.0.3',
          'dendropy>=4.5.2',
          'ete3>=3.1.1',
      ],
      extras_require={
          'webserver': [
              'dash==1.21.0',
              'dash-bio==0.8.0',
              'dash-bootstrap-components==0.12.2',
              'Flask==2.0.1',
              'gunicorn==20.1.0',
              'plotly==5.1.0',
              'supervisor==4.2.2',
              'rq==1.10.0',
              'hiredis==2.0.0',
          ]
      }
      )



#!/usr/bin/env python
import os
import uuid
import numpy as np
import numpy as np
import re

from matplotlib import pyplot as plt
from matplotlib.patches import PathPatch,Rectangle,Circle,Polygon
from matplotlib.path import Path
import mpl_toolkits.mplot3d.art3d as art3d
from matplotlib.text import TextPath
from matplotlib.transforms import Affine2D

from .pholy import auto_detect_groups




def read_file(filename, filetype):

    seq_dict = {}
    seqnames = []
    seqname = None
    seq = None
    ith = 0
    if filetype.lower() in ['fasta','fa']:
        with open(filename,'r') as inpf:
            for line in inpf:
                line = line.strip()
                if len(line)==0:
                    continue
                if line[0] == '>':
                    ith += 1
                    seqname = f'{line[1:]} {ith}'
                    seqnames.append(seqname)
                else:
                    seq_dict[seqname]  = seq_dict.get(seqname,'') + line.upper()
    elif filetype.lower() in ['fastq','fq']:
        with open(filename,'r') as inpf:
            num = -1
            for line in inpf:
                line = line.strip()
                if len(line)==0:
                    continue
                num += 1
                if num%4 == 0:
                    assert line[0] == '@'
                    seqname = f"{line[1:]} {num+1}"
                    seqnames.append(seqname)
                if num%4 == 1:
                    seq_dict[seqname] = line.upper()
    else:
        pass

    return seq_dict, seqnames

def grouping(seqs,seq_file='',sequence_type='aa',group_by='length',group_resolution=1,clustering_method='max',
             clustalo_bin='', fasttree_bin='',fasttreemp_bin='', treecluster_bin='',
             uid='',fa_output_dir='.',figure_output_dir='.'):
    
    groups_dict = {}
    if group_by.lower() == 'length':
        for name,seq in seqs:
            key = f'Len{len(seq)}'
            if key not in groups_dict:
                groups_dict[key] = []
            groups_dict[key].append([name,seq])
    elif group_by.lower() == 'identifier':
        for name,seq in seqs:
            group_pat = re.search('group@(\d+-\S+)',name)
            if group_pat:
                group_id = group_pat.groups()[0]
                if group_id not in groups_dict:
                    groups_dict[group_id] = []
                groups_dict[group_id].append([name,seq])
    elif group_by.lower() == 'auto':
        groups_dict = auto_detect_groups(seqs,seq_file,sequence_type,group_resolution,clustering_method,
                                         clustalo_bin,fasttree_bin,fasttreemp_bin,treecluster_bin,
                                         uid,fa_output_dir,figure_output_dir)
    
    return groups_dict

def check_group(groups):
    for group_id in groups:
        seqs = groups[group_id]
        if len(set([len(x[1]) for x in seqs])) > 1:
            print('Sequence lengths not same in one group')
            exit(0)

def write_to_tmp(seqs,tmp_path = './tmp/'):
    if not os.path.exists(tmp_path):
        os.mkdir(tmp_path)

    uid = str(uuid.uuid4())
    f_name = os.path.join(tmp_path,f'{uid}.fa')
    with open(f_name,'w') as outpf:
        for name,seq in seqs:
            outpf.write(f'>{name}\n')
            outpf.write(f'{seq}\n')
    return f_name


def angle_between(p1, p2=(0,0)):
    ang1 = np.arctan2(*p1[::-1])
    ang2 = np.arctan2(*p2[::-1])
    return (ang1 - ang2) % (2 * np.pi)

def get_coor_by_angle(radius, angle, origin=(0,0)):
    relative_coor =  (radius * np.cos(angle), radius * np.sin(angle))
    return (relative_coor[0]+origin[0],relative_coor[1]+origin[1])


def check_parallel(edge1, edge2):
    start1, end1 = edge1
    start2, end2 = edge2
    shifted_end1 = (end1[0]-start1[0], end1[1]-start1[1])
    shifted_end2 = (end2[0]-start2[0], end2[1]-start2[1])

def curve_connect(leftbot,lefttop,righttop,rightbot,limit_width,direction='right',**kargs):
    if np.abs((leftbot[0] - lefttop[0])/limit_width)< 0.1:
        return Polygon(xy=[leftbot,lefttop,righttop,rightbot], **kargs)
    else:
        if direction == 'left':
            limit_width = -limit_width
        p0 = leftbot
        p1 = (leftbot[0] + limit_width, leftbot[1])
        p2 = (lefttop[0] + limit_width, lefttop[1])
        p3 = lefttop
        p4 = righttop
        p5 = (righttop[0] + limit_width,righttop[1])
        p6 = (rightbot[0] + limit_width, rightbot[1])
        p7 = rightbot
        p8 = p0

        verts = [p0,p1,p2,p3,p4,p5,p6,p7,p8]

        codes = [
            Path.MOVETO,
            Path.CURVE4,
            Path.CURVE4,
            Path.CURVE4,
            Path.LINETO,
            Path.CURVE4,
            Path.CURVE4,
            Path.CURVE4,
            Path.CLOSEPOLY
        ]
        path = Path(verts, codes)
        patch = PathPatch(path, **kargs)
        return patch

def straight_connect(p1,p2,p3,p4,**kargs):
    verts = [p1,p2,p3,p4,p1]
    codes = [
        Path.MOVETO,
        Path.LINETO,
        Path.LINETO,
        Path.LINETO,
        Path.CLOSEPOLY
    ]
    path = Path(verts, codes)
    patch = PathPatch(path, **kargs)
    return patch

def link_edges(edge1, edge2, ax, threed=False,x=0,y=1,z=-1, color='blue',alpha=0.1):
    if ax is None:
        _, ax = plt.subplots(1, 1,figsize=(10,10))
    
    
    p1,p2 = edge1
    p4,p3 = edge2

    if threed:
        patch = straight_connect((p1[x],p1[y]),
                                 (p2[x],p2[y]),
                                 (p3[x],p3[y]),
                                 (p4[x],p4[y]), fill=True,alpha=alpha,color=color,linewidth=0)

    else:
        patch = straight_connect(p1,p2,p3,p4, fill=True,alpha=alpha,color=color,linewidth=0)


    ax.add_patch(patch)
    if threed:
        art3d.pathpatch_2d_to_3d(patch, z=0, zdir='z')

    return ax


#https://stackoverflow.com/questions/34372480/rotate-point-about-another-point-in-degrees-python/34374437
def rotate(p, origin=(0, 0), angle=0):
    R = np.array([[np.cos(angle), -np.sin(angle)],
                  [np.sin(angle),  np.cos(angle)]])
    o = np.atleast_2d(origin)
    p = np.atleast_2d(p)
    return np.squeeze((R @ (p.T-o.T) + o.T).T)



#https://matplotlib.org/stable/gallery/mplot3d/pathpatch3d.html#sphx-glr-gallery-mplot3d-pathpatch3d-py
def text3d(ax, xyz, s, zdir="z", size=None, angle=0, usetex=False, **kwargs):
    """
    Plots the string *s* on the axes *ax*, with position *xyz*, size *size*,
    and rotation angle *angle*. *zdir* gives the axis which is to be treated as
    the third dimension. *usetex* is a boolean indicating whether the string
    should be run through a LaTeX subprocess or not.  Any additional keyword
    arguments are forwarded to `.transform_path`.

    Note: zdir affects the interpretation of xyz.
    """
    x, y, z = xyz
    if zdir == "y":
        xy1, z1 = (x, z), y
    elif zdir == "x":
        xy1, z1 = (y, z), x
    else:
        xy1, z1 = (x, y), z

    text_path = TextPath((0, 0), s, size=size, usetex=usetex)
    trans = Affine2D().rotate(angle).translate(xy1[0], xy1[1])

    p1 = PathPatch(trans.transform_path(text_path), **kwargs)
    ax.add_patch(p1)
    art3d.pathpatch_2d_to_3d(p1, z=z1, zdir=zdir)

def detect_seq_type(seqs):
    

    dna_set = {'A','T','G','C','N','-'}
    rna_set = {'A','U','G','C','N','-'}
    protein_set = {'A','R','N','D','C','Q','E','G','H','I','L','K','M','F','P','S','T','W','Y','V','-'}

    base_set = set()
    for _,seq in seqs:
        base_set |= set(seq)


    seq_type = 'aa'
    if base_set.issubset(dna_set):
        seq_type = 'dna' 
    elif base_set.issubset(rna_set):
        seq_type = 'rna' 
    elif base_set.issubset(protein_set):
        seq_type = 'aa' 
    
    return seq_type
    
def save_seqs(seqs, filename):
    with open(filename,'w') as outpf:
        for seqname,seq in seqs:
            outpf.write(f'>{seqname}\n')
            outpf.write(f'{seq}\n')

